import datetime

from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Handshake(_message.Message):
    __slots__ = ("sot",)
    SOT_FIELD_NUMBER: _ClassVar[int]
    sot: bool
    def __init__(self, sot: bool = ...) -> None: ...

class SourceTransformRequest(_message.Message):
    __slots__ = ("request", "handshake")
    class Request(_message.Message):
        __slots__ = ("keys", "value", "event_time", "watermark", "headers", "id")
        class HeadersEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        KEYS_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        EVENT_TIME_FIELD_NUMBER: _ClassVar[int]
        WATERMARK_FIELD_NUMBER: _ClassVar[int]
        HEADERS_FIELD_NUMBER: _ClassVar[int]
        ID_FIELD_NUMBER: _ClassVar[int]
        keys: _containers.RepeatedScalarFieldContainer[str]
        value: bytes
        event_time: _timestamp_pb2.Timestamp
        watermark: _timestamp_pb2.Timestamp
        headers: _containers.ScalarMap[str, str]
        id: str
        def __init__(self, keys: _Optional[_Iterable[str]] = ..., value: _Optional[bytes] = ..., event_time: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., watermark: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., headers: _Optional[_Mapping[str, str]] = ..., id: _Optional[str] = ...) -> None: ...
    REQUEST_FIELD_NUMBER: _ClassVar[int]
    HANDSHAKE_FIELD_NUMBER: _ClassVar[int]
    request: SourceTransformRequest.Request
    handshake: Handshake
    def __init__(self, request: _Optional[_Union[SourceTransformRequest.Request, _Mapping]] = ..., handshake: _Optional[_Union[Handshake, _Mapping]] = ...) -> None: ...

class SourceTransformResponse(_message.Message):
    __slots__ = ("results", "id", "handshake")
    class Result(_message.Message):
        __slots__ = ("keys", "value", "event_time", "tags")
        KEYS_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        EVENT_TIME_FIELD_NUMBER: _ClassVar[int]
        TAGS_FIELD_NUMBER: _ClassVar[int]
        keys: _containers.RepeatedScalarFieldContainer[str]
        value: bytes
        event_time: _timestamp_pb2.Timestamp
        tags: _containers.RepeatedScalarFieldContainer[str]
        def __init__(self, keys: _Optional[_Iterable[str]] = ..., value: _Optional[bytes] = ..., event_time: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., tags: _Optional[_Iterable[str]] = ...) -> None: ...
    RESULTS_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    HANDSHAKE_FIELD_NUMBER: _ClassVar[int]
    results: _containers.RepeatedCompositeFieldContainer[SourceTransformResponse.Result]
    id: str
    handshake: Handshake
    def __init__(self, results: _Optional[_Iterable[_Union[SourceTransformResponse.Result, _Mapping]]] = ..., id: _Optional[str] = ..., handshake: _Optional[_Union[Handshake, _Mapping]] = ...) -> None: ...

class ReadyResponse(_message.Message):
    __slots__ = ("ready",)
    READY_FIELD_NUMBER: _ClassVar[int]
    ready: bool
    def __init__(self, ready: bool = ...) -> None: ...



import threading
from concurrent.futures import ThreadPoolExecutor
from collections.abc import Iterable

from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2

from pynumaflow.shared.server import exit_on_error
from pynumaflow.shared.synciter import SyncIterator
from pynumaflow.sourcetransformer import Datum
from pynumaflow.sourcetransformer._dtypes import SourceTransformCallable
from pynumaflow.proto.sourcetransformer import transform_pb2
from pynumaflow.proto.sourcetransformer import transform_pb2_grpc
from pynumaflow.types import NumaflowServicerContext
from pynumaflow._constants import (
    _LOGGER,
    STREAM_EOF,
    NUM_THREADS_DEFAULT,
    ERR_UDF_EXCEPTION_STRING,
)


def _create_read_handshake_response() -> transform_pb2.SourceTransformResponse:
    """
    Create a handshake response for the SourceTransform function.

    Returns:
    SourceTransformResponse: A SourceTransformResponse object indicating a successful handshake.
    """
    return transform_pb2.SourceTransformResponse(
        handshake=transform_pb2.Handshake(sot=True),
    )


class SourceTransformServicer(transform_pb2_grpc.SourceTransformServicer):
    """
    This class is used to create a new grpc SourceTransform servicer instance.
    It implements the SourceTransformServicer interface from the proto transform.proto file.
    Provides the functionality for the required rpc methods.
    """

    def __init__(self, handler: SourceTransformCallable, multiproc: bool = False):
        self.__transform_handler: SourceTransformCallable = handler
        # This indicates whether the grpc server attached is multiproc or not
        self.multiproc = multiproc
        # create a thread pool for executing UDF code
        self.executor = ThreadPoolExecutor(max_workers=NUM_THREADS_DEFAULT)

    def SourceTransformFn(
        self,
        request_iterator: Iterable[transform_pb2.SourceTransformRequest],
        context: NumaflowServicerContext,
    ) -> Iterable[transform_pb2.SourceTransformResponse]:
        """
        Applies a function to each datum element.
        The pascal case function name comes from the generated transform_pb2_grpc.py file.
        """
        try:
            # The first message to be received should be a valid handshake
            req = next(request_iterator)
            # check if it is a valid handshake req
            if not (req.handshake and req.handshake.sot):
                raise Exception("SourceTransformFn: expected handshake message")
            yield _create_read_handshake_response()

            # result queue to stream messages from the user code back to the client
            result_queue = SyncIterator()

            # Reader thread to keep reading from the request iterator and schedule
            # execution for each of them
            reader_thread = threading.Thread(
                target=self._process_requests, args=(context, request_iterator, result_queue)
            )
            reader_thread.start()

            # Read the result queue and keep forwarding them upstream
            for res in result_queue.read_iterator():
                # if error handler accordingly
                if isinstance(res, BaseException):
                    # Terminate the current server process due to exception
                    exit_on_error(
                        context, f"{ERR_UDF_EXCEPTION_STRING}: {repr(res)}", parent=self.multiproc
                    )
                    return
                # return the result
                yield res

            # wait for the threads to clean-up
            reader_thread.join()
            self.executor.shutdown(cancel_futures=True)

        except BaseException as err:
            _LOGGER.critical("UDFError, re-raising the error", exc_info=True)
            # Terminate the current server process due to exception
            exit_on_error(
                context, f"{ERR_UDF_EXCEPTION_STRING}: {repr(err)}", parent=self.multiproc
            )
            return

    def _process_requests(
        self,
        context: NumaflowServicerContext,
        request_iterator: Iterable[transform_pb2.SourceTransformRequest],
        result_queue: SyncIterator,
    ):
        try:
            # read through all incoming requests and submit to the
            # threadpool for invocation
            for request in request_iterator:
                _ = self.executor.submit(self._invoke_transformer, context, request, result_queue)
            # wait for all tasks to finish after all requests exhausted
            self.executor.shutdown(wait=True)
            # Indicate to the result queue that no more messages left to process
            result_queue.put(STREAM_EOF)
        except BaseException:
            _LOGGER.critical("SourceTransformFnError, re-raising the error", exc_info=True)

    def _invoke_transformer(
        self, context, request: transform_pb2.SourceTransformRequest, result_queue: SyncIterator
    ):
        try:
            d = Datum(
                keys=list(request.request.keys),
                value=request.request.value,
                event_time=request.request.event_time.ToDatetime(),
                watermark=request.request.watermark.ToDatetime(),
                headers=dict(request.request.headers),
            )
            responses = self.__transform_handler(list(request.request.keys), d)

            results = []
            for resp in responses:
                event_time_timestamp = _timestamp_pb2.Timestamp()
                event_time_timestamp.FromDatetime(dt=resp.event_time)
                results.append(
                    transform_pb2.SourceTransformResponse.Result(
                        keys=list(resp.keys),
                        value=resp.value,
                        tags=resp.tags,
                        event_time=event_time_timestamp,
                    )
                )
            result_queue.put(
                transform_pb2.SourceTransformResponse(results=results, id=request.request.id)
            )
        except BaseException as e:
            _LOGGER.critical("SourceTransform handler error", exc_info=True)
            result_queue.put(e)
            return

    def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> transform_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto transform_pb2_grpc.py file.
        """
        return transform_pb2.ReadyResponse(ready=True)



from google.protobuf import empty_pb2 as _empty_pb2

from pynumaflow._constants import (
    _LOGGER,
    ERR_UDF_EXCEPTION_STRING,
)
from pynumaflow.proto.sideinput import sideinput_pb2_grpc, sideinput_pb2
from pynumaflow.shared.server import exit_on_error
from pynumaflow.sideinput._dtypes import RetrieverCallable
from pynumaflow.types import NumaflowServicerContext


class SideInputServicer(sideinput_pb2_grpc.SideInputServicer):
    def __init__(
        self,
        handler: RetrieverCallable,
    ):
        self.__retrieve_handler: RetrieverCallable = handler

    def RetrieveSideInput(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> sideinput_pb2.SideInputResponse:
        """
        Applies a sideinput function for a retrieval request.
        The pascal case function name comes from the proto sideinput_pb2_grpc.py file.
        """
        # if there is an exception, we will mark all the responses as a failure
        try:
            rspn = self.__retrieve_handler()
        except BaseException as err:
            err_msg = f"{ERR_UDF_EXCEPTION_STRING}: {repr(err)}"
            _LOGGER.critical(err_msg, exc_info=True)
            exit_on_error(context, err_msg)
            return

        return sideinput_pb2.SideInputResponse(value=rspn.value, no_broadcast=rspn.no_broadcast)

    def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> sideinput_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto sideinput_pb2_grpc.py file.
        """
        return sideinput_pb2.ReadyResponse(ready=True)



import asyncio
from collections.abc import AsyncIterable

from google.protobuf import empty_pb2 as _empty_pb2
from pynumaflow.shared.asynciter import NonBlockingIterator

from pynumaflow._constants import _LOGGER, STREAM_EOF, ERR_UDF_EXCEPTION_STRING
from pynumaflow.mapper._dtypes import MapAsyncCallable, Datum, MapError
from pynumaflow.proto.mapper import map_pb2, map_pb2_grpc
from pynumaflow.shared.server import handle_async_error
from pynumaflow.types import NumaflowServicerContext


class AsyncMapServicer(map_pb2_grpc.MapServicer):
    """
    This class is used to create a new grpc Async Map Servicer instance.
    It implements the SyncMapServicer interface from the proto map.proto file.
    Provides the functionality for the required rpc methods.
    """

    def __init__(
        self,
        handler: MapAsyncCallable,
    ):
        self.background_tasks = set()
        self.__map_handler: MapAsyncCallable = handler

    async def MapFn(
        self,
        request_iterator: AsyncIterable[map_pb2.MapRequest],
        context: NumaflowServicerContext,
    ) -> AsyncIterable[map_pb2.MapResponse]:
        """
        Applies a function to each datum element.
        The pascal case function name comes from the proto map_pb2_grpc.py file.
        """
        # proto repeated field(keys) is of type google._upb._message.RepeatedScalarContainer
        # we need to explicitly convert it to list
        try:
            # The first message to be received should be a valid handshake
            req = await request_iterator.__anext__()
            # check if it is a valid handshake req
            if not (req.handshake and req.handshake.sot):
                raise MapError("MapFn: expected handshake as the first message")
            yield map_pb2.MapResponse(handshake=map_pb2.Handshake(sot=True))

            global_result_queue = NonBlockingIterator()

            # reader task to process the input task and invoke the required tasks
            producer = asyncio.create_task(
                self._process_inputs(request_iterator, global_result_queue)
            )

            # keep reading on result queue and send messages back
            consumer = global_result_queue.read_iterator()
            async for msg in consumer:
                # If the message is an exception, we raise the exception
                if isinstance(msg, BaseException):
                    await handle_async_error(context, msg, ERR_UDF_EXCEPTION_STRING)
                    return
                # Send window response back to the client
                else:
                    yield msg
            # wait for the producer task to complete
            await producer
        except BaseException as e:
            _LOGGER.critical("UDFError, re-raising the error", exc_info=True)
            await handle_async_error(context, e, ERR_UDF_EXCEPTION_STRING)
            return

    async def _process_inputs(
        self,
        request_iterator: AsyncIterable[map_pb2.MapRequest],
        result_queue: NonBlockingIterator,
    ):
        """
        Utility function for processing incoming MapRequests
        """
        try:
            # for each incoming request, create a background task to execute the
            # UDF code
            async for req in request_iterator:
                msg_task = asyncio.create_task(self._invoke_map(req, result_queue))
                # save a reference to a set to store active tasks
                self.background_tasks.add(msg_task)
                msg_task.add_done_callback(self.background_tasks.discard)

            # wait for all tasks to complete
            for task in self.background_tasks:
                await task

            # send an EOF to result queue to indicate that all tasks have completed
            await result_queue.put(STREAM_EOF)

        except BaseException:
            _LOGGER.critical("MapFn Error, re-raising the error", exc_info=True)

    async def _invoke_map(self, req: map_pb2.MapRequest, result_queue: NonBlockingIterator):
        """
        Invokes the user defined function.
        """
        try:
            datum = Datum(
                keys=list(req.request.keys),
                value=req.request.value,
                event_time=req.request.event_time.ToDatetime(),
                watermark=req.request.watermark.ToDatetime(),
                headers=dict(req.request.headers),
            )
            msgs = await self.__map_handler(list(req.request.keys), datum)
            datums = []
            for msg in msgs:
                datums.append(
                    map_pb2.MapResponse.Result(keys=msg.keys, value=msg.value, tags=msg.tags)
                )
            await result_queue.put(map_pb2.MapResponse(results=datums, id=req.id))
        except BaseException as err:
            _LOGGER.critical("MapFn handler error", exc_info=True)
            await result_queue.put(err)

    async def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> map_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto map_pb2_grpc.py file.
        """
        return map_pb2.ReadyResponse(ready=True)



import asyncio
from datetime import datetime, timezone
from typing import Union
from collections.abc import AsyncIterable

from pynumaflow.exceptions import UDFError
from pynumaflow.proto.reducer import reduce_pb2
from pynumaflow.shared.asynciter import NonBlockingIterator
from pynumaflow._constants import (
    STREAM_EOF,
    DELIMITER,
    _LOGGER,
)
from pynumaflow.reducestreamer._dtypes import (
    IntervalWindow,
    ReduceResult,
    Datum,
    _ReduceStreamBuilderClass,
    ReduceStreamAsyncCallable,
    ReduceWindow,
    WindowOperation,
    Metadata,
)


def build_unique_key_name(keys, window):
    """
    Builds a unique key name for the given keys and window.
    The key name is used to identify the Reduce task.
    The format is: start_time:end_time:key1:key2:...
    """
    return f"{window.start.ToMilliseconds()}:{window.end.ToMilliseconds()}:{DELIMITER.join(keys)}"


def build_window_hash(window):
    """
    Builds a hash for the given window.
    The hash is used to identify the Reduce Window
    The format is: start_time:end_time
    """
    return f"{window.start.ToMilliseconds()}:{window.end.ToMilliseconds()}"


def create_window_eof_response(window):
    """Create a Reduce response with EOF=True for a given window"""
    return reduce_pb2.ReduceResponse(window=window, EOF=True)


class TaskManager:
    """
    TaskManager is responsible for managing the Reduce tasks.
    It is created whenever a new reduce operation is requested.
    """

    def __init__(self, handler: Union[ReduceStreamAsyncCallable, _ReduceStreamBuilderClass]):
        # A dictionary to store the task information
        self.tasks = {}
        # Collection for storing strong references to all running tasks.
        # Event loop only keeps a weak reference, which can cause it to
        # get lost during execution.
        self.background_tasks = set()
        # Handler for the reduce operation
        self.__reduce_handler = handler
        # Queue to store the results of the reduce operation
        # This queue is used to send the results to the client
        # once the reduce operation is completed.
        # This queue is also used to send the error/exceptions to the client
        # if the reduce operation fails.
        self.global_result_queue = NonBlockingIterator()

    def get_unique_windows(self):
        """
        Returns the unique windows that are currently being processed
        """
        # Dict to store unique windows
        windows = dict()
        # Iterate over all the tasks and add the windows
        for task in self.tasks.values():
            window_hash = build_window_hash(task.window)
            window_found = windows.get(window_hash, None)
            # if window not seen yet, add to the dict
            if not window_found:
                windows[window_hash] = task.window
        return windows

    def get_tasks(self):
        """
        Returns the list of reduce tasks that are
        currently being processed
        """
        return list(self.tasks.values())

    async def stream_send_eof(self):
        """
        Sends EOF to input streams of all the Reduce
        tasks that are currently being processed.
        This is called when the input grpc stream is closed.
        """
        for unified_key in self.tasks:
            await self.tasks[unified_key].iterator.put(STREAM_EOF)

    async def create_task(self, req):
        """
        Creates a new reduce task for the given request.
        Based on the request we compute a unique key, and then
        it creates a new task or appends the request to the existing task.
        """
        # if len of windows in request != 1, raise error
        if len(req.windows) != 1:
            raise UDFError("reduce create operation error: invalid number of windows")

        d = req.payload
        keys = d.keys
        unified_key = build_unique_key_name(keys, req.windows[0])
        curr_task = self.tasks.get(unified_key, None)

        # If the task does not exist, create a new task
        if not curr_task:
            niter = NonBlockingIterator()
            riter = niter.read_iterator()
            # Create a new result queue for the current task
            # We create a new result queue for each task, so that
            # the results of the reduce operation can be sent to the
            # the global result queue, which in turn sends the results
            # to the client.
            res_queue = NonBlockingIterator()

            # Create a new write_to_global_queue task for the current, this will read from the
            # result queue specifically for the current task and update the
            # global result queue
            consumer = asyncio.create_task(
                self.write_to_global_queue(res_queue, self.global_result_queue, req.windows[0])
            )
            # Save a reference to the result of this function, to avoid a
            # task disappearing mid-execution.
            self.background_tasks.add(consumer)
            consumer.add_done_callback(self.clean_background)

            # Create a new task for the reduce operation, this will invoke the
            # Reduce handler with the given keys, request iterator, and window.
            task = asyncio.create_task(self.__invoke_reduce(keys, riter, res_queue, req.windows[0]))
            # Save a reference to the result of this function, to avoid a
            # task disappearing mid-execution.
            self.background_tasks.add(task)
            task.add_done_callback(self.clean_background)

            # Create a new ReduceResult object to store the task information
            curr_task = ReduceResult(task, niter, keys, req.windows[0], res_queue, consumer)

            # Save the result of the reduce operation to the task list
            self.tasks[unified_key] = curr_task

        # Put the request in the iterator
        await curr_task.iterator.put(d)

    async def send_datum_to_task(self, req):
        """
        Appends the request to the existing window reduce task.
        If the task does not exist, create it.
        """
        if len(req.windows) != 1:
            raise UDFError("reduce append operation error: invalid number of windows")
        d = req.payload
        keys = d.keys
        unified_key = build_unique_key_name(keys, req.windows[0])
        result = self.tasks.get(unified_key, None)
        if not result:
            await self.create_task(req)
        else:
            await result.iterator.put(d)

    async def __invoke_reduce(
        self,
        keys: list[str],
        request_iterator: AsyncIterable[Datum],
        output: NonBlockingIterator,
        window: ReduceWindow,
    ):
        """
        Invokes the UDF reduce handler with the given keys,
        request iterator, and window. Returns the result of the
        reduce operation.
        """
        new_instance = self.__reduce_handler

        # Convert the window to a datetime object
        start_dt = datetime.fromtimestamp(int(window.start.ToMilliseconds()) / 1e3, timezone.utc)
        end_dt = datetime.fromtimestamp(int(window.end.ToMilliseconds()) / 1e3, timezone.utc)
        interval_window = IntervalWindow(start_dt, end_dt)
        md = Metadata(interval_window=interval_window)
        # If the reduce handler is a class instance, create a new instance of it.
        # It is required for a new key to be processed by a
        # new instance of the reducer for a given window
        # Otherwise the function handler can be called directly
        if isinstance(self.__reduce_handler, _ReduceStreamBuilderClass):
            new_instance = self.__reduce_handler.create()
        try:
            _ = await new_instance(keys, request_iterator, output, md)
        # If there is an error in the reduce operation, log and
        # then send the error to the result queue
        except BaseException as err:
            _LOGGER.critical("UDFError, re-raising the error", exc_info=True)
            # Put the exception in the result queue
            await self.global_result_queue.put(err)

    async def process_input_stream(self, request_iterator: AsyncIterable[reduce_pb2.ReduceRequest]):
        # Start iterating through the request iterator and create tasks
        # based on the operation type received.
        try:
            async for request in request_iterator:
                # check whether the request is an open or append operation
                if request.operation is int(WindowOperation.OPEN):
                    # create a new task for the open operation and
                    # put the request in the task iterator
                    await self.create_task(request)
                elif request.operation is int(WindowOperation.APPEND):
                    # append the task data to the existing task
                    # if the task does not exist, create a new task
                    await self.send_datum_to_task(request)
        # If there is an error in the reduce operation, log and
        # then send the error to the result queue
        except BaseException as e:
            err_msg = f"Reduce Streaming Error: {repr(e)}"
            _LOGGER.critical(err_msg, exc_info=True)
            # Put the exception in the global result queue
            await self.global_result_queue.put(e)
            return

        try:
            # send EOF to all the tasks once the request iterator is exhausted
            # This will signal the tasks to stop reading the data on their
            # respective iterators.
            await self.stream_send_eof()

            # get the list of reduce tasks that are currently being processed
            # iterate through the tasks and wait for them to complete
            for task in self.get_tasks():
                # Once this is done, we know that the task has written all the results
                # to the local result queue
                fut = task.future
                await fut

                # Send an EOF message to the local result queue
                # This will signal that the task has completed processing
                await task.result_queue.put(STREAM_EOF)

                # Wait for the local queue to write
                # all the results of this task to the global result queue
                con_future = task.consumer_future
                await con_future

            # Once all tasks are completed, senf EOF to all windows that
            # were processed in the Task Manager. We send a single
            # EOF message per window.
            current_windows = self.get_unique_windows()
            for window in current_windows.values():
                # Send an EOF message to the global result queue
                # This will signal that window has been processed
                eof_window_msg = create_window_eof_response(window=window)
                await self.global_result_queue.put(eof_window_msg)

            # Once all tasks are completed, senf EOF the global result queue
            await self.global_result_queue.put(STREAM_EOF)
        except BaseException as e:
            err_msg = f"Reduce Streaming Error: {repr(e)}"
            _LOGGER.critical(err_msg, exc_info=True)
            await self.global_result_queue.put(e)

    async def write_to_global_queue(
        self, input_queue: NonBlockingIterator, output_queue: NonBlockingIterator, window
    ):
        """
        This task is for given Reduce task.
        This would from the local result queue for the task and then write
        to the global result queue
        """
        reader = input_queue.read_iterator()
        async for msg in reader:
            res = reduce_pb2.ReduceResponse.Result(keys=msg.keys, value=msg.value, tags=msg.tags)
            out = reduce_pb2.ReduceResponse(result=res, window=window)
            await output_queue.put(out)

    def clean_background(self, task):
        """
        Remove the task from the background tasks collection
        """
        self.background_tasks.remove(task)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: pynumaflow/proto/sideinput/sideinput.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'pynumaflow/proto/sideinput/sideinput.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*pynumaflow/proto/sideinput/sideinput.proto\x12\x0csideinput.v1\x1a\x1bgoogle/protobuf/empty.proto\"8\n\x11SideInputResponse\x12\r\n\x05value\x18\x01 \x01(\x0c\x12\x14\n\x0cno_broadcast\x18\x02 \x01(\x08\"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08\x32\x99\x01\n\tSideInput\x12L\n\x11RetrieveSideInput\x12\x16.google.protobuf.Empty\x1a\x1f.sideinput.v1.SideInputResponse\x12>\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a\x1b.sideinput.v1.ReadyResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pynumaflow.proto.sideinput.sideinput_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SIDEINPUTRESPONSE']._serialized_start=89
  _globals['_SIDEINPUTRESPONSE']._serialized_end=145
  _globals['_READYRESPONSE']._serialized_start=147
  _globals['_READYRESPONSE']._serialized_end=177
  _globals['_SIDEINPUT']._serialized_start=180
  _globals['_SIDEINPUT']._serialized_end=333
# @@protoc_insertion_point(module_scope)



import datetime

from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Payload(_message.Message):
    __slots__ = ("keys", "value", "event_time", "watermark", "id", "headers")
    class HeadersEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    KEYS_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    EVENT_TIME_FIELD_NUMBER: _ClassVar[int]
    WATERMARK_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    HEADERS_FIELD_NUMBER: _ClassVar[int]
    keys: _containers.RepeatedScalarFieldContainer[str]
    value: bytes
    event_time: _timestamp_pb2.Timestamp
    watermark: _timestamp_pb2.Timestamp
    id: str
    headers: _containers.ScalarMap[str, str]
    def __init__(self, keys: _Optional[_Iterable[str]] = ..., value: _Optional[bytes] = ..., event_time: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., watermark: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., id: _Optional[str] = ..., headers: _Optional[_Mapping[str, str]] = ...) -> None: ...

class AccumulatorRequest(_message.Message):
    __slots__ = ("payload", "operation")
    class WindowOperation(_message.Message):
        __slots__ = ("event", "keyedWindow")
        class Event(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            OPEN: _ClassVar[AccumulatorRequest.WindowOperation.Event]
            CLOSE: _ClassVar[AccumulatorRequest.WindowOperation.Event]
            APPEND: _ClassVar[AccumulatorRequest.WindowOperation.Event]
        OPEN: AccumulatorRequest.WindowOperation.Event
        CLOSE: AccumulatorRequest.WindowOperation.Event
        APPEND: AccumulatorRequest.WindowOperation.Event
        EVENT_FIELD_NUMBER: _ClassVar[int]
        KEYEDWINDOW_FIELD_NUMBER: _ClassVar[int]
        event: AccumulatorRequest.WindowOperation.Event
        keyedWindow: KeyedWindow
        def __init__(self, event: _Optional[_Union[AccumulatorRequest.WindowOperation.Event, str]] = ..., keyedWindow: _Optional[_Union[KeyedWindow, _Mapping]] = ...) -> None: ...
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    payload: Payload
    operation: AccumulatorRequest.WindowOperation
    def __init__(self, payload: _Optional[_Union[Payload, _Mapping]] = ..., operation: _Optional[_Union[AccumulatorRequest.WindowOperation, _Mapping]] = ...) -> None: ...

class KeyedWindow(_message.Message):
    __slots__ = ("start", "end", "slot", "keys")
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    SLOT_FIELD_NUMBER: _ClassVar[int]
    KEYS_FIELD_NUMBER: _ClassVar[int]
    start: _timestamp_pb2.Timestamp
    end: _timestamp_pb2.Timestamp
    slot: str
    keys: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, start: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., end: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., slot: _Optional[str] = ..., keys: _Optional[_Iterable[str]] = ...) -> None: ...

class AccumulatorResponse(_message.Message):
    __slots__ = ("payload", "window", "tags", "EOF")
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    WINDOW_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    EOF_FIELD_NUMBER: _ClassVar[int]
    payload: Payload
    window: KeyedWindow
    tags: _containers.RepeatedScalarFieldContainer[str]
    EOF: bool
    def __init__(self, payload: _Optional[_Union[Payload, _Mapping]] = ..., window: _Optional[_Union[KeyedWindow, _Mapping]] = ..., tags: _Optional[_Iterable[str]] = ..., EOF: bool = ...) -> None: ...

class ReadyResponse(_message.Message):
    __slots__ = ("ready",)
    READY_FIELD_NUMBER: _ClassVar[int]
    ready: bool
    def __init__(self, ready: bool = ...) -> None: ...



from pynumaflow.mapper.async_server import MapAsyncServer
from pynumaflow.mapper.multiproc_server import MapMultiprocServer
from pynumaflow.mapper.sync_server import MapServer

from pynumaflow.mapper._dtypes import Message, Messages, Datum, DROP, Mapper

__all__ = [
    "Message",
    "Messages",
    "Datum",
    "DROP",
    "Mapper",
    "MapServer",
    "MapAsyncServer",
    "MapMultiprocServer",
]



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from pynumaflow.proto.sourcer import source_pb2 as pynumaflow_dot_proto_dot_sourcer_dot_source__pb2

GRPC_GENERATED_VERSION = '1.75.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in pynumaflow/proto/sourcer/source_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class SourceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ReadFn = channel.stream_stream(
                '/source.v1.Source/ReadFn',
                request_serializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.ReadRequest.SerializeToString,
                response_deserializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.ReadResponse.FromString,
                _registered_method=True)
        self.AckFn = channel.stream_stream(
                '/source.v1.Source/AckFn',
                request_serializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.AckRequest.SerializeToString,
                response_deserializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.AckResponse.FromString,
                _registered_method=True)
        self.NackFn = channel.unary_unary(
                '/source.v1.Source/NackFn',
                request_serializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.NackRequest.SerializeToString,
                response_deserializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.NackResponse.FromString,
                _registered_method=True)
        self.PendingFn = channel.unary_unary(
                '/source.v1.Source/PendingFn',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.PendingResponse.FromString,
                _registered_method=True)
        self.PartitionsFn = channel.unary_unary(
                '/source.v1.Source/PartitionsFn',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.PartitionsResponse.FromString,
                _registered_method=True)
        self.IsReady = channel.unary_unary(
                '/source.v1.Source/IsReady',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.ReadyResponse.FromString,
                _registered_method=True)


class SourceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ReadFn(self, request_iterator, context):
        """Read returns a stream of datum responses.
        The size of the returned responses is less than or equal to the num_records specified in each ReadRequest.
        If the request timeout is reached on the server side, the returned responses will contain all the datum that have been read (which could be an empty list).
        The server will continue to read and respond to subsequent ReadRequests until the client closes the stream.
        Once it has sent all the datum, the server will send a ReadResponse with the end of transmission flag set to true.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AckFn(self, request_iterator, context):
        """AckFn acknowledges a stream of datum offsets.
        When AckFn is called, it implicitly indicates that the datum stream has been processed by the source vertex.
        The caller (numa) expects the AckFn to be successful, and it does not expect any errors.
        If there are some irrecoverable errors when the callee (UDSource) is processing the AckFn request,
        then it is best to crash because there are no other retry mechanisms possible.
        Clients sends n requests and expects n responses.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def NackFn(self, request, context):
        """NackFn negatively acknowledges a batch of offsets. Invoked during a critical error in the monovertex or pipeline.
        Unlike AckFn its not a streaming rpc because this is only invoked when there is a critical error (error path).
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PendingFn(self, request, context):
        """PendingFn returns the number of pending records at the user defined source.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PartitionsFn(self, request, context):
        """PartitionsFn returns the list of partitions for the user defined source.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IsReady(self, request, context):
        """IsReady is the heartbeat endpoint for user defined source gRPC.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SourceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ReadFn': grpc.stream_stream_rpc_method_handler(
                    servicer.ReadFn,
                    request_deserializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.ReadRequest.FromString,
                    response_serializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.ReadResponse.SerializeToString,
            ),
            'AckFn': grpc.stream_stream_rpc_method_handler(
                    servicer.AckFn,
                    request_deserializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.AckRequest.FromString,
                    response_serializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.AckResponse.SerializeToString,
            ),
            'NackFn': grpc.unary_unary_rpc_method_handler(
                    servicer.NackFn,
                    request_deserializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.NackRequest.FromString,
                    response_serializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.NackResponse.SerializeToString,
            ),
            'PendingFn': grpc.unary_unary_rpc_method_handler(
                    servicer.PendingFn,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.PendingResponse.SerializeToString,
            ),
            'PartitionsFn': grpc.unary_unary_rpc_method_handler(
                    servicer.PartitionsFn,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.PartitionsResponse.SerializeToString,
            ),
            'IsReady': grpc.unary_unary_rpc_method_handler(
                    servicer.IsReady,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.ReadyResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'source.v1.Source', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('source.v1.Source', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Source(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ReadFn(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/source.v1.Source/ReadFn',
            pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.ReadRequest.SerializeToString,
            pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.ReadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AckFn(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/source.v1.Source/AckFn',
            pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.AckRequest.SerializeToString,
            pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.AckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def NackFn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/source.v1.Source/NackFn',
            pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.NackRequest.SerializeToString,
            pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.NackResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def PendingFn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/source.v1.Source/PendingFn',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.PendingResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def PartitionsFn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/source.v1.Source/PartitionsFn',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.PartitionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def IsReady(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/source.v1.Source/IsReady',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            pynumaflow_dot_proto_dot_sourcer_dot_source__pb2.ReadyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from pynumaflow._constants import (
    NUM_THREADS_DEFAULT,
    MAX_MESSAGE_SIZE,
    MAP_SOCK_PATH,
    UDFType,
    _PROCESS_COUNT,
    MAP_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)
from pynumaflow.info.server import get_metadata_env
from pynumaflow.info.types import (
    ServerInfo,
    METADATA_ENVS,
    MAP_MODE_KEY,
    MapMode,
    MULTIPROC_KEY,
    MINIMUM_NUMAFLOW_VERSION,
    ContainerType,
)
from pynumaflow.mapper._dtypes import MapSyncCallable
from pynumaflow.mapper._servicer._sync_servicer import SyncMapServicer
from pynumaflow.shared.server import (
    NumaflowServer,
    start_multiproc_server,
)


class MapMultiprocServer(NumaflowServer):
    """
    Create a new grpc Multiproc Map Server instance.
    """

    def __init__(
        self,
        mapper_instance: MapSyncCallable,
        server_count: int = _PROCESS_COUNT,
        sock_path=MAP_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=MAP_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Multiproc Map Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
            mapper_instance: The mapper instance to be used for Map UDF
            server_count: The number of grpc server instances to be forked for multiproc
            sock_path: The UNIX socket path to be used for the server
            max_message_size: The max message size in bytes the server can receive and send
            max_threads: The max number of threads to be spawned;
                        defaults to 4 and max capped at 16

        Example invocation:
            import math
            import os
            from pynumaflow.mapper import Messages, Message, Datum, Mapper, MapMultiprocServer

            def is_prime(n):
                for i in range(2, int(math.ceil(math.sqrt(n)))):
                    if n % i == 0:
                        return False
                else:
                    return True

            class PrimeMap(Mapper):
                def handler(self, keys: list[str], datum: Datum) -> Messages:
                    val = datum.value
                    _ = datum.event_time
                    _ = datum.watermark
                    messages = Messages()
                    for i in range(2, 100000):
                        is_prime(i)
                    messages.append(Message(val, keys=keys))
                    return messages

            if __name__ == "__main__":
                server_count = 2
                prime_class = PrimeMap()
                # Server count is the number of server processes to start
                grpc_server = MapMultiprocServer(prime_class, server_count=server_count)
                grpc_server.start()

        """
        self.sock_path = f"unix://{sock_path}"
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.max_message_size = max_message_size
        self.server_info_file = server_info_file

        self.mapper_instance = mapper_instance

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
            ("grpc.so_reuseport", 1),
            ("grpc.so_reuseaddr", 1),
        ]
        # Set the number of processes to be spawned to the number of CPUs or
        # the value of the parameter server_count defined by the user
        # Setting the max value to 2 * CPU count
        # Used for multiproc server
        self._process_count = min(server_count, 2 * _PROCESS_COUNT)
        self.servicer = SyncMapServicer(handler=mapper_instance, multiproc=True)

    def start(self) -> None:
        """
        Starts the N grpc servers gRPC serves on the with
        given max threads.
        where N = The number of CPUs or the value of the parameter server_count
        defined by the user. The max value is capped to 2 * CPU count.
        """

        # Create the server info file
        server_info = ServerInfo.get_default_server_info()
        server_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[ContainerType.Mapper]
        server_info.metadata = get_metadata_env(envs=METADATA_ENVS)
        # Add the MULTIPROC metadata using the number of servers to use
        server_info.metadata[MULTIPROC_KEY] = str(self._process_count)
        # Add the MAP_MODE metadata to the server info for the correct map mode
        server_info.metadata[MAP_MODE_KEY] = MapMode.UnaryMap

        # Start the multiproc server
        start_multiproc_server(
            max_threads=self.max_threads,
            servicer=self.servicer,
            process_count=self._process_count,
            server_info_file=self.server_info_file,
            server_options=self._server_options,
            udf_type=UDFType.Map,
            server_info=server_info,
        )



from abc import ABCMeta, abstractmethod
from asyncio import Task
from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from typing import TypeVar, Callable, Union, Optional
from collections.abc import AsyncIterable

from pynumaflow.shared.asynciter import NonBlockingIterator
from pynumaflow._constants import DROP

M = TypeVar("M", bound="Message")


class WindowOperation(IntEnum):
    """
    Enumerate the type of Window operation received.
    """

    OPEN = (0,)
    CLOSE = (1,)
    APPEND = (2,)


@dataclass(init=False)
class Datum:
    """
    Class to define the important information for the event.
    Args:
        keys: the keys of the event.
        value: the payload of the event.
        event_time: the event time of the event.
        watermark: the watermark of the event.
    >>> # Example usage
    >>> from pynumaflow.accumulator import Datum
    >>> from datetime import datetime, timezone
    >>> payload = bytes("test_mock_message", encoding="utf-8")
    >>> t1 = datetime.fromtimestamp(1662998400, timezone.utc)
    >>> t2 = datetime.fromtimestamp(1662998460, timezone.utc)
    >>> msg_headers = {"key1": "value1", "key2": "value2"}
    >>> d = Datum(
    ...       keys=["test_key"],
    ...       value=payload,
    ...       event_time=t1,
    ...       watermark=t2,
    ...       headers=msg_headers
    ...    )
    """

    __slots__ = ("_keys", "_value", "_event_time", "_watermark", "_headers", "_id")

    _keys: list[str]
    _value: bytes
    _event_time: datetime
    _watermark: datetime
    _headers: dict[str, str]
    _id: str

    def __init__(
        self,
        keys: list[str],
        value: bytes,
        event_time: datetime,
        watermark: datetime,
        id_: str,
        headers: Optional[dict[str, str]] = None,
    ):
        self._keys = keys or list()
        self._value = value or b""
        if not isinstance(event_time, datetime):
            raise TypeError(f"Wrong data type: {type(event_time)} for Datum.event_time")
        self._event_time = event_time
        if not isinstance(watermark, datetime):
            raise TypeError(f"Wrong data type: {type(watermark)} for Datum.watermark")
        self._watermark = watermark
        self._headers = headers or {}
        self._id = id_

    @property
    def keys(self) -> list[str]:
        """Returns the keys of the event.

        Returns:
            list[str]: A list of string keys associated with this event.
        """
        return self._keys

    @property
    def value(self) -> bytes:
        """Returns the value of the event.

        Returns:
            bytes: The payload data of the event as bytes.
        """
        return self._value

    @property
    def event_time(self) -> datetime:
        """Returns the event time of the event.

        Returns:
            datetime: The timestamp when the event occurred.
        """
        return self._event_time

    @property
    def watermark(self) -> datetime:
        """Returns the watermark of the event.

        Returns:
            datetime: The watermark timestamp indicating the progress of event time.
        """
        return self._watermark

    @property
    def headers(self) -> dict[str, str]:
        """Returns the headers of the event.

        Returns:
            dict[str, str]: A dictionary containing header key-value pairs for this event.
        """
        return self._headers.copy()

    @property
    def id(self) -> str:
        """Returns the id of the event.

        Returns:
            str: The unique identifier for this event.
        """
        return self._id


@dataclass(init=False)
class IntervalWindow:
    """Defines the start and end of the interval window for the event."""

    __slots__ = ("_start", "_end")

    _start: datetime
    _end: datetime

    def __init__(self, start: datetime, end: datetime):
        self._start = start
        self._end = end

    @property
    def start(self) -> datetime:
        """Returns the start point of the interval window.

        Returns:
            datetime: The start timestamp of the interval window.
        """
        return self._start

    @property
    def end(self) -> datetime:
        """Returns the end point of the interval window.

        Returns:
            datetime: The end timestamp of the interval window.
        """
        return self._end


@dataclass(init=False)
class KeyedWindow:
    """
    Defines the window for a accumulator operation which includes the
    interval window along with the slot.
    """

    __slots__ = ("_window", "_slot", "_keys")

    _window: IntervalWindow
    _slot: str
    _keys: list[str]

    def __init__(self, start: datetime, end: datetime, slot: str = "", keys: list[str] = []):
        self._window = IntervalWindow(start=start, end=end)
        self._slot = slot
        self._keys = keys

    @property
    def start(self) -> datetime:
        """Returns the start point of the interval window.

        Returns:
            datetime: The start timestamp of the interval window.
        """
        return self._window.start

    @property
    def end(self) -> datetime:
        """Returns the end point of the interval window.

        Returns:
            datetime: The end timestamp of the interval window.
        """
        return self._window.end

    @property
    def slot(self) -> str:
        """Returns the slot from the window.

        Returns:
            str: The slot identifier for this window.
        """
        return self._slot

    @property
    def window(self) -> IntervalWindow:
        """Returns the interval window.

        Returns:
            IntervalWindow: The underlying interval window object.
        """
        return self._window

    @property
    def keys(self) -> list[str]:
        """Returns the keys for window.

        Returns:
            list[str]: A list of keys associated with this window.
        """
        return self._keys


@dataclass
class AccumulatorResult:
    """Defines the object to hold the result of accumulator computation."""

    __slots__ = (
        "_future",
        "_iterator",
        "_key",
        "_result_queue",
        "_consumer_future",
        "_latest_watermark",
    )

    _future: Task
    _iterator: NonBlockingIterator
    _key: list[str]
    _result_queue: NonBlockingIterator
    _consumer_future: Task
    _latest_watermark: datetime

    @property
    def future(self) -> Task:
        """Returns the future result of computation.

        Returns:
            Task: The asyncio Task representing the computation future.
        """
        return self._future

    @property
    def iterator(self) -> NonBlockingIterator:
        """Returns the handle to the producer queue.

        Returns:
            NonBlockingIterator: The iterator for producing data to the queue.
        """
        return self._iterator

    @property
    def keys(self) -> list[str]:
        """Returns the keys of the partition.

        Returns:
            list[str]: The keys associated with this partition.
        """
        return self._key

    @property
    def result_queue(self) -> NonBlockingIterator:
        """Returns the async queue used to write the output for the tasks.

        Returns:
            NonBlockingIterator: The queue for writing task output.
        """
        return self._result_queue

    @property
    def consumer_future(self) -> Task:
        """Returns the async consumer task for the result queue.

        Returns:
            Task: The asyncio Task for consuming from the result queue.
        """
        return self._consumer_future

    @property
    def latest_watermark(self) -> datetime:
        """Returns the latest watermark for task.

        Returns:
            datetime: The latest watermark timestamp for this task.
        """
        return self._latest_watermark

    def update_watermark(self, new_watermark: datetime):
        """Updates the latest watermark value.

        Args:
            new_watermark (datetime): The new watermark timestamp to set.

        Raises:
            TypeError: If new_watermark is not a datetime object.
        """
        if not isinstance(new_watermark, datetime):
            raise TypeError("new_watermark must be a datetime object")
        self._latest_watermark = new_watermark


@dataclass
class AccumulatorRequest:
    """Defines the object to hold a request for the accumulator operation."""

    __slots__ = ("_operation", "_keyed_window", "_payload")

    _operation: WindowOperation
    _keyed_window: KeyedWindow
    _payload: Datum

    def __init__(self, operation: WindowOperation, keyed_window: KeyedWindow, payload: Datum):
        self._operation = operation
        self._keyed_window = keyed_window
        self._payload = payload

    @property
    def operation(self) -> WindowOperation:
        """Returns the operation type.

        Returns:
            WindowOperation: The type of window operation (OPEN, CLOSE, or APPEND).
        """
        return self._operation

    @property
    def keyed_window(self) -> KeyedWindow:
        """Returns the keyed window.

        Returns:
            KeyedWindow: The keyed window associated with this request.
        """
        return self._keyed_window

    @property
    def payload(self) -> Datum:
        """Returns the payload of the window.

        Returns:
            Datum: The data payload for this accumulator request.
        """
        return self._payload


@dataclass(init=False)
class Message:
    """
    Basic datatype for data passing to the next vertex/vertices.

    Args:
        value: data in bytes
        keys: []string keys for vertex (optional)
        tags: []string tags for conditional forwarding (optional)
        watermark: watermark for this message (optional)
        event_time: event time for this message (optional)
        headers: headers for this message (optional)
        id: message id (optional)
    """

    __slots__ = ("_value", "_keys", "_tags", "_watermark", "_event_time", "_headers", "_id")

    _value: bytes
    _keys: list[str]
    _tags: list[str]
    _watermark: datetime
    _event_time: datetime
    _headers: dict[str, str]
    _id: str

    def __init__(
        self,
        value: bytes,
        keys: list[str] = None,
        tags: list[str] = None,
        watermark: datetime = None,
        event_time: datetime = None,
        headers: dict[str, str] = None,
        id: str = None,
    ):
        """
        Creates a Message object to send value to a vertex.
        """
        self._keys = keys or []
        self._tags = tags or []
        self._value = value or b""
        self._watermark = watermark
        self._event_time = event_time
        self._headers = headers or {}
        self._id = id or ""

    @classmethod
    def to_drop(cls: type[M]) -> M:
        """Creates a Message instance that indicates the message should be dropped.

        Returns:
            M: A Message instance with empty value and DROP tag indicating
               the message should be dropped.
        """
        return cls(b"", None, [DROP])

    @property
    def value(self) -> bytes:
        """Returns the message payload value.

        Returns:
            bytes: The message payload data as bytes.
        """
        return self._value

    @property
    def keys(self) -> list[str]:
        """Returns the message keys.

        Returns:
            list[str]: A list of string keys associated with this message.
        """
        return self._keys

    @property
    def tags(self) -> list[str]:
        """Returns the message tags for conditional forwarding.

        Returns:
            list[str]: A list of string tags used for conditional forwarding.
        """
        return self._tags

    @property
    def watermark(self) -> datetime:
        """Returns the watermark timestamp for this message.

        Returns:
            datetime: The watermark timestamp, or None if not set.
        """
        return self._watermark

    @property
    def event_time(self) -> datetime:
        """Returns the event time for this message.

        Returns:
            datetime: The event time timestamp, or None if not set.
        """
        return self._event_time

    @property
    def headers(self) -> dict[str, str]:
        """Returns the message headers.

        Returns:
            dict[str, str]: A dictionary containing header key-value pairs for this message.
        """
        return self._headers.copy()

    @property
    def id(self) -> str:
        """Returns the message ID.

        Returns:
            str: The unique identifier for this message.
        """
        return self._id

    @classmethod
    def from_datum(cls, datum: Datum):
        """Create a Message instance from a Datum object.

        Args:
            datum: The Datum object to convert

        Returns:
            Message: A new Message instance with data from the datum
        """
        return cls(
            value=datum.value,
            keys=datum.keys,
            watermark=datum.watermark,
            event_time=datum.event_time,
            headers=datum.headers,
            id=datum.id,
        )


AccumulatorAsyncCallable = Callable[[list[str], AsyncIterable[Datum], NonBlockingIterator], None]


class Accumulator(metaclass=ABCMeta):
    """
    Accumulate can read unordered from the input stream and emit the ordered
    data to the output stream. Once the watermark (WM) of the output stream progresses,
    the data in WAL until that WM will be garbage collected.
    NOTE: A message can be silently dropped if need be,
    and it will be cleared from the WAL when the WM progresses.
    """

    def __call__(self, *args, **kwargs):
        """
        Allow to call handler function directly if class instance is sent
        as the accumulator_instance.
        """
        return self.handler(*args, **kwargs)

    @abstractmethod
    async def handler(
        self,
        datums: AsyncIterable[Datum],
        output: NonBlockingIterator,
    ):
        """
        Implement this handler function which implements the AccumulatorStreamCallable interface.
        """
        pass


class _AccumulatorBuilderClass:
    """
    Class to build an Accumulator class instance.
    Used Internally

    Args:
        accumulator_class: the Accumulator class to be used for Accumulator UDF
        args: the arguments to be passed to the accumulator class
        kwargs: the keyword arguments to be passed to the accumulator class
    """

    def __init__(self, accumulator_class: type[Accumulator], args: tuple, kwargs: dict):
        self._accumulator_class: type[Accumulator] = accumulator_class
        self._args = args
        self._kwargs = kwargs

    def create(self) -> Accumulator:
        """
        Create a new Accumulator instance.
        """
        return self._accumulator_class(*self._args, **self._kwargs)


# AccumulatorStreamCallable is a callable which can be used as a handler for the Accumulator UDF.
AccumulatorStreamCallable = Union[AccumulatorAsyncCallable, type[Accumulator]]



from pynumaflow.info.types import ContainerType, MINIMUM_NUMAFLOW_VERSION, ServerInfo
from pynumaflow._constants import (
    MAX_MESSAGE_SIZE,
    SOURCE_TRANSFORMER_SOCK_PATH,
    NUM_THREADS_DEFAULT,
    _LOGGER,
    UDFType,
    SOURCE_TRANSFORMER_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)
from pynumaflow.shared import NumaflowServer
from pynumaflow.shared.server import sync_server_start
from pynumaflow.sourcetransformer._dtypes import SourceTransformCallable
from pynumaflow.sourcetransformer.servicer._servicer import SourceTransformServicer


class SourceTransformServer(NumaflowServer):
    """
    Class for a new Source Transformer Server instance.
    """

    def __init__(
        self,
        source_transform_instance: SourceTransformCallable,
        sock_path=SOURCE_TRANSFORMER_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=SOURCE_TRANSFORMER_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Source Transformer Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
            source_transform_instance: The source transformer instance to be used for
            Source Transformer UDF
            sock_path: The UNIX socket path to be used for the server
            max_message_size: The max message size in bytes the server can receive and send
            max_threads: The max number of threads to be spawned;
                            defaults to 4 and max capped at 16

        Example Invocation:

        import datetime
        import logging

        from pynumaflow.sourcetransformer import Messages, Message, Datum, SourceTransformServer
        # This is a simple User Defined Function example which receives a message,
        # applies the following
        # data transformation, and returns the message.
        # If the message event time is before year 2022, drop the message with event time unchanged.
        # If it's within year 2022, update the tag to "within_year_2022" and
        # update the message event time to Jan 1st 2022.
        # Otherwise, (exclusively after year 2022), update the tag to
        # "after_year_2022" and update the
        # message event time to Jan 1st 2023.

        january_first_2022 = datetime.datetime.fromtimestamp(1640995200)
        january_first_2023 = datetime.datetime.fromtimestamp(1672531200)


        def my_handler(keys: list[str], datum: Datum) -> Messages:
            val = datum.value
            event_time = datum.event_time
            messages = Messages()

            if event_time < january_first_2022:
                logging.info("Got event time:%s, it is before 2022, so dropping", event_time)
                messages.append(Message.to_drop(event_time))
            elif event_time < january_first_2023:
                logging.info(
                    "Got event time:%s, it is within year 2022, so forwarding to within_year_2022",
                    event_time,
                )
                messages.append(
                    Message(value=val, event_time=january_first_2022,
                            tags=["within_year_2022"])
                )
            else:
                logging.info(
                    "Got event time:%s, it is after year 2022, so forwarding to
                    after_year_2022", event_time
                )
                messages.append(Message(value=val, event_time=january_first_2023,
                                tags=["after_year_2022"]))

            return messages


        if __name__ == "__main__":
            grpc_server = SourceTransformServer(my_handler)
            grpc_server.start()
        """
        self.sock_path = f"unix://{sock_path}"
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.max_message_size = max_message_size
        self.server_info_file = server_info_file

        self.source_transform_instance = source_transform_instance

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]
        self.servicer = SourceTransformServicer(handler=source_transform_instance)

    def start(self):
        """
        Starts the Synchronous gRPC server on the given UNIX socket with given max threads.
        """
        _LOGGER.info(
            "Sync GRPC Server listening on: %s with max threads: %s",
            self.sock_path,
            self.max_threads,
        )
        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[
            ContainerType.Sourcetransformer
        ]
        # Start the sync server
        sync_server_start(
            servicer=self.servicer,
            bind_address=self.sock_path,
            max_threads=self.max_threads,
            server_info_file=self.server_info_file,
            server_options=self._server_options,
            udf_type=UDFType.SourceTransformer,
            server_info=serv_info,
        )



import asyncio
from collections.abc import AsyncIterable

from google.protobuf import empty_pb2 as _empty_pb2

from pynumaflow.shared.asynciter import NonBlockingIterator
from pynumaflow._constants import _LOGGER, STREAM_EOF, ERR_UDF_EXCEPTION_STRING
from pynumaflow.mapstreamer import Datum
from pynumaflow.mapstreamer._dtypes import MapStreamCallable, MapStreamError
from pynumaflow.proto.mapper import map_pb2_grpc, map_pb2
from pynumaflow.shared.server import handle_async_error
from pynumaflow.types import NumaflowServicerContext


class AsyncMapStreamServicer(map_pb2_grpc.MapServicer):
    """
    Concurrent gRPC Map Stream Servicer.
    Spawns one background task per incoming MapRequest; each task streams
    results as produced and finally emits an EOT for that request.
    """

    def __init__(self, handler: MapStreamCallable):
        self.__map_stream_handler: MapStreamCallable = handler
        self._background_tasks: set[asyncio.Task] = set()

    async def MapFn(
        self,
        request_iterator: AsyncIterable[map_pb2.MapRequest],
        context: NumaflowServicerContext,
    ) -> AsyncIterable[map_pb2.MapResponse]:
        """
        Applies a map function to a datum stream in streaming mode.
        The PascalCase name comes from the generated map_pb2_grpc.py file.
        """
        try:
            # First message must be a handshake
            first = await request_iterator.__anext__()
            if not (first.handshake and first.handshake.sot):
                raise MapStreamError("MapStreamFn: expected handshake as the first message")
            # Acknowledge handshake
            yield map_pb2.MapResponse(handshake=map_pb2.Handshake(sot=True))

            # Global non-blocking queue for outbound responses / errors
            global_result_queue = NonBlockingIterator()

            # Start producer that turns each inbound request into a background task
            producer = asyncio.create_task(
                self._process_inputs(request_iterator, global_result_queue)
            )

            # Consume results as they arrive and stream them to the client
            async for msg in global_result_queue.read_iterator():
                if isinstance(msg, BaseException):
                    await handle_async_error(context, msg, ERR_UDF_EXCEPTION_STRING)
                    return
                else:
                    # msg is a map_pb2.MapResponse, already formed
                    yield msg

            # Ensure producer has finished (covers graceful shutdown)
            await producer

        except BaseException as e:
            _LOGGER.critical("UDFError, re-raising the error", exc_info=True)
            await handle_async_error(context, e, ERR_UDF_EXCEPTION_STRING)
            return

    async def _process_inputs(
        self,
        request_iterator: AsyncIterable[map_pb2.MapRequest],
        result_queue: NonBlockingIterator,
    ) -> None:
        """
        Reads MapRequests from the client and spawns a background task per request.
        Each task streams results to result_queue as they are produced.
        """
        try:
            async for req in request_iterator:
                task = asyncio.create_task(self._invoke_map_stream(req, result_queue))
                self._background_tasks.add(task)
                # Remove from the set when done to avoid memory growth
                task.add_done_callback(self._background_tasks.discard)

            # Wait for all in-flight tasks to complete
            if self._background_tasks:
                await asyncio.gather(*list(self._background_tasks), return_exceptions=False)

            # Signal end-of-stream to the consumer
            await result_queue.put(STREAM_EOF)

        except BaseException as e:
            _LOGGER.critical("MapFn Error, re-raising the error", exc_info=True)
            # Surface the error to the consumer; MapFn will handle and close the RPC
            await result_queue.put(e)

    async def _invoke_map_stream(
        self,
        req: map_pb2.MapRequest,
        result_queue: NonBlockingIterator,
    ) -> None:
        """
        Invokes the user-provided async generator for a single request and
        pushes each result onto the global queue, followed by an EOT for this id.
        """
        try:
            datum = Datum(
                keys=list(req.request.keys),
                value=req.request.value,
                event_time=req.request.event_time.ToDatetime(),
                watermark=req.request.watermark.ToDatetime(),
                headers=dict(req.request.headers),
            )

            # Stream results from the user handler as they are produced
            async for msg in self.__map_stream_handler(list(req.request.keys), datum):
                res = map_pb2.MapResponse.Result(keys=msg.keys, value=msg.value, tags=msg.tags)
                await result_queue.put(map_pb2.MapResponse(results=[res], id=req.id))

            # Emit EOT for this request id
            await result_queue.put(
                map_pb2.MapResponse(status=map_pb2.TransmissionStatus(eot=True), id=req.id)
            )

        except BaseException as err:
            _LOGGER.critical("MapFn handler error", exc_info=True)
            # Surface handler error to the main producer;
            # it will call handle_async_error and end the RPC
            await result_queue.put(err)

    async def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> map_pb2.ReadyResponse:
        """Heartbeat endpoint for gRPC."""
        return map_pb2.ReadyResponse(ready=True)



import aiorun
import grpc

from pynumaflow.info.types import (
    ServerInfo,
    MAP_MODE_KEY,
    MapMode,
    MINIMUM_NUMAFLOW_VERSION,
    ContainerType,
)
from pynumaflow.mapstreamer.servicer.async_servicer import AsyncMapStreamServicer
from pynumaflow.proto.mapper import map_pb2_grpc

from pynumaflow._constants import (
    MAP_STREAM_SOCK_PATH,
    MAX_MESSAGE_SIZE,
    NUM_THREADS_DEFAULT,
    _LOGGER,
    MAP_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)

from pynumaflow.mapstreamer._dtypes import MapStreamCallable

from pynumaflow.shared.server import NumaflowServer, start_async_server


class MapStreamAsyncServer(NumaflowServer):
    """
    Class for a new Map Stream Server instance.
    """

    def __init__(
        self,
        map_stream_instance: MapStreamCallable,
        sock_path=MAP_STREAM_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=MAP_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Async Map Stream Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
            map_stream_instance: The map stream instance to be used for Map Stream UDF
            sock_path: The UNIX socket path to be used for the server
            max_message_size: The max message size in bytes the server can receive and send
            max_threads: The max number of threads to be spawned;
                            defaults to 4 and max capped at 16
            server_type: The type of server to be used

        Example invocation:
            import os
            from collections.abc import AsyncIterable
            from pynumaflow.mapstreamer import Message, Datum, MapStreamAsyncServer, MapStreamer

            class FlatMapStream(MapStreamer):
                async def handler(self, keys: list[str], datum: Datum) -> AsyncIterable[Message]:
                    val = datum.value
                    _ = datum.event_time
                    _ = datum.watermark
                    strs = val.decode("utf-8").split(",")

                    if len(strs) == 0:
                        yield Message.to_drop()
                        return
                    for s in strs:
                        yield Message(str.encode(s))

            async def map_stream_handler(_: list[str], datum: Datum) -> AsyncIterable[Message]:

                val = datum.value
                _ = datum.event_time
                _ = datum.watermark
                strs = val.decode("utf-8").split(",")

                if len(strs) == 0:
                    yield Message.to_drop()
                    return
                for s in strs:
                    yield Message(str.encode(s))

            if __name__ == "__main__":
                invoke = os.getenv("INVOKE", "func_handler")
                if invoke == "class":
                    handler = FlatMapStream()
                else:
                    handler = map_stream_handler
                grpc_server = MapStreamAsyncServer(handler)
                grpc_server.start()

        """
        self.map_stream_instance: MapStreamCallable = map_stream_instance
        self.sock_path = f"unix://{sock_path}"
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.max_message_size = max_message_size
        self.server_info_file = server_info_file

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]

        self.servicer = AsyncMapStreamServicer(handler=self.map_stream_instance)

    def start(self):
        """
        Starter function for the Async Map Stream server, we need a separate caller
        to the aexec so that all the async coroutines can be started from a single context
        """
        aiorun.run(self.aexec(), use_uvloop=True)

    async def aexec(self):
        """
        Starts the Async gRPC server on the given UNIX socket with
        given max threads.
        """
        # As the server is async, we need to create a new server instance in the
        # same thread as the event loop so that all the async calls are made in the
        # same context
        # Create a new async server instance and add the servicer to it
        server = grpc.aio.server(options=self._server_options)
        server.add_insecure_port(self.sock_path)
        map_pb2_grpc.add_MapServicer_to_server(
            self.servicer,
            server,
        )
        _LOGGER.info("Starting Map Stream Server")
        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[ContainerType.Mapper]
        # Add the MAP_MODE metadata to the server info for the correct map mode
        serv_info.metadata[MAP_MODE_KEY] = MapMode.StreamMap

        # Start the async server
        await start_async_server(
            server_async=server,
            sock_path=self.sock_path,
            max_threads=self.max_threads,
            cleanup_coroutines=list(),
            server_info_file=self.server_info_file,
            server_info=serv_info,
        )



from pynumaflow.sinker.async_server import SinkAsyncServer

from pynumaflow.sinker.server import SinkServer

from pynumaflow.sinker._dtypes import Response, Responses, Datum, Sinker

__all__ = ["Response", "Responses", "Datum", "Sinker", "SinkAsyncServer", "SinkServer"]



import os

from pynumaflow.info.types import ServerInfo, ContainerType, MINIMUM_NUMAFLOW_VERSION
from pynumaflow.sinker.servicer.sync_servicer import SyncSinkServicer

from pynumaflow._constants import (
    SINK_SOCK_PATH,
    MAX_MESSAGE_SIZE,
    NUM_THREADS_DEFAULT,
    _LOGGER,
    UDFType,
    SINK_SERVER_INFO_FILE_PATH,
    ENV_UD_CONTAINER_TYPE,
    UD_CONTAINER_FALLBACK_SINK,
    FALLBACK_SINK_SOCK_PATH,
    FALLBACK_SINK_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)

from pynumaflow.shared.server import NumaflowServer, sync_server_start
from pynumaflow.sinker._dtypes import SinkSyncCallable


class SinkServer(NumaflowServer):
    """
    SinkServer is the main class to start a gRPC server for a sinker.
    """

    def __init__(
        self,
        sinker_instance: SinkSyncCallable,
        sock_path=SINK_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=SINK_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Sink Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
            sinker_instance: The sinker instance to be used for Sink UDF
            sock_path: The UNIX socket path to be used for the server
            max_message_size: The max message size in bytes the server can receive and send
            max_threads: The max number of threads to be spawned;
                            defaults to 4 and max capped at 16
        Example invocation:
            import os
            from collections.abc import Iterator

            from pynumaflow.sinker import Datum, Responses, Response, SinkServer
            from pynumaflow.sinker import Sinker
            from pynumaflow._constants import _LOGGER

            class UserDefinedSink(Sinker):
                def handler(self, datums: Iterator[Datum]) -> Responses:
                    responses = Responses()
                    for msg in datums:
                        _LOGGER.info("User Defined Sink %s", msg.value.decode("utf-8"))
                        responses.append(Response.as_success(msg.id))
                    return responses

            def udsink_handler(datums: Iterator[Datum]) -> Responses:
                responses = Responses()
                for msg in datums:
                    _LOGGER.info("User Defined Sink %s", msg.value.decode("utf-8"))
                    responses.append(Response.as_success(msg.id))
                return responses

            if __name__ == "__main__":
                invoke = os.getenv("INVOKE", "func_handler")
                if invoke == "class":
                    sink_handler = UserDefinedSink()
                else:
                    sink_handler = udsink_handler
                grpc_server = SinkServer(sink_handler)
                grpc_server.start()

        """
        # If the container type is fallback sink, then use the fallback sink address and path.
        if os.getenv(ENV_UD_CONTAINER_TYPE, "") == UD_CONTAINER_FALLBACK_SINK:
            _LOGGER.info("Using Fallback Sink")
            sock_path = FALLBACK_SINK_SOCK_PATH
            server_info_file = FALLBACK_SINK_SERVER_INFO_FILE_PATH

        self.sock_path = f"unix://{sock_path}"
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.max_message_size = max_message_size
        self.server_info_file = server_info_file

        self.sinker_instance = sinker_instance

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]
        self.servicer = SyncSinkServicer(sinker_instance)

    def start(self):
        """
        Starts the Synchronous gRPC server on the
        given UNIX socket with given max threads.
        """
        _LOGGER.info(
            "Sync GRPC Sink listening on: %s with max threads: %s",
            self.sock_path,
            self.max_threads,
        )
        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[ContainerType.Sinker]
        # Start the server
        sync_server_start(
            servicer=self.servicer,
            bind_address=self.sock_path,
            max_threads=self.max_threads,
            server_info_file=self.server_info_file,
            server_options=self._server_options,
            udf_type=UDFType.Sink,
            server_info=serv_info,
        )



from pynumaflow._constants import SIDE_INPUT_DIR_PATH
from pynumaflow.sideinput._dtypes import Response, SideInput
from pynumaflow.sideinput.server import SideInputServer

__all__ = ["Response", "SideInput", "SideInputServer", "SIDE_INPUT_DIR_PATH"]



import asyncio
from collections.abc import AsyncIterable

from google.protobuf import empty_pb2 as _empty_pb2

from pynumaflow.batchmapper import Datum
from pynumaflow.batchmapper._dtypes import BatchMapCallable, BatchMapError
from pynumaflow.proto.mapper import map_pb2, map_pb2_grpc
from pynumaflow.shared.asynciter import NonBlockingIterator
from pynumaflow.shared.server import handle_async_error
from pynumaflow.types import NumaflowServicerContext
from pynumaflow._constants import _LOGGER, STREAM_EOF, ERR_UDF_EXCEPTION_STRING


class AsyncBatchMapServicer(map_pb2_grpc.MapServicer):
    """
    This class is used to create a new grpc Batch Map Servicer instance.
    It implements the MapServicer interface from the proto
    map_pb2_grpc.py file.
    Provides the functionality for the required rpc methods.
    """

    def __init__(
        self,
        handler: BatchMapCallable,
    ):
        self.background_tasks = set()
        self.__batch_map_handler: BatchMapCallable = handler

    async def MapFn(
        self,
        request_iterator: AsyncIterable[map_pb2.MapRequest],
        context: NumaflowServicerContext,
    ) -> AsyncIterable[map_pb2.MapResponse]:
        """
        Applies a batch map function to a MapRequest stream in a batching mode.
        The pascal case function name comes from the proto map_pb2_grpc.py file.
        """
        try:
            # The first message to be received should be a valid handshake
            req = await request_iterator.__anext__()
            # check if it is a valid handshake req
            if not (req.handshake and req.handshake.sot):
                raise BatchMapError("BatchMapFn: expected handshake as the first message")
            yield map_pb2.MapResponse(handshake=map_pb2.Handshake(sot=True))

            # cur_task is used to track the task (coroutine) processing
            # the current batch of messages.
            cur_task = None
            # iterate of the incoming messages ot the sink
            async for d in request_iterator:
                # if we do not have any active task currently processing the batch
                # we need to create one and call the User function for processing the same.
                if cur_task is None:
                    req_queue = NonBlockingIterator()
                    cur_task = asyncio.create_task(
                        self.__batch_map_handler(req_queue.read_iterator())
                    )
                    self.background_tasks.add(cur_task)
                    cur_task.add_done_callback(self.background_tasks.discard)
                # when we have end of transmission message, we need to stop the processing the
                # current batch and wait for the next batch of messages.
                # We will also wait for the current task to finish processing the current batch.
                # We mark the current task as None to indicate that we are
                # ready to process the next batch.
                if d.status and d.status.eot:
                    await req_queue.put(STREAM_EOF)
                    await cur_task
                    ret = cur_task.result()

                    # iterate over the responses received and covert to the required proto format
                    for batch_response in ret:
                        single_req_resp = []
                        for msg in batch_response.messages:
                            single_req_resp.append(
                                map_pb2.MapResponse.Result(
                                    keys=msg.keys, value=msg.value, tags=msg.tags
                                )
                            )
                        # send the response for a given ID back to the stream
                        yield map_pb2.MapResponse(id=batch_response.id, results=single_req_resp)

                    # send EOT after each finishing Batch responses
                    yield map_pb2.MapResponse(status=map_pb2.TransmissionStatus(eot=True))
                    cur_task = None
                    continue

                # if we have a valid message, we will add it to the request queue for processing.
                datum = Datum(
                    keys=list(d.request.keys),
                    value=d.request.value,
                    event_time=d.request.event_time.ToDatetime(),
                    watermark=d.request.watermark.ToDatetime(),
                    headers=dict(d.request.headers),
                    id=d.id,
                )
                await req_queue.put(datum)

        except BaseException as err:
            _LOGGER.critical("UDFError, re-raising the error", exc_info=True)
            await handle_async_error(context, err, ERR_UDF_EXCEPTION_STRING)
            return

    async def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> map_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto batchmap_pb2_grpc.py file.
        """
        return map_pb2.ReadyResponse(ready=True)



import aiorun
import grpc

from pynumaflow._constants import (
    NUM_THREADS_DEFAULT,
    MAX_MESSAGE_SIZE,
    MAX_NUM_THREADS,
    SOURCE_TRANSFORMER_SOCK_PATH,
    SOURCE_TRANSFORMER_SERVER_INFO_FILE_PATH,
)
from pynumaflow.info.types import (
    ServerInfo,
    MINIMUM_NUMAFLOW_VERSION,
    ContainerType,
)
from pynumaflow.proto.sourcetransformer import transform_pb2_grpc
from pynumaflow.shared.server import (
    NumaflowServer,
    start_async_server,
)
from pynumaflow.sourcetransformer._dtypes import SourceTransformAsyncCallable
from pynumaflow.sourcetransformer.servicer._async_servicer import SourceTransformAsyncServicer


class SourceTransformAsyncServer(NumaflowServer):
    """
    Create a new grpc Source Transformer Server instance.
    A new servicer instance is created and attached to the server.
    The server instance is returned.
    Args:
        source_transform_instance: The source transformer instance to be used for
        Source Transformer UDF
        sock_path: The UNIX socket path to be used for the server
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                        defaults to 4 and max capped at 16

    Example Invocation:

    import datetime
    import logging

    from pynumaflow.sourcetransformer import Messages, Message, Datum, SourceTransformServer
    # This is a simple User Defined Function example which receives a message,
    # applies the following
    # data transformation, and returns the message.
    # If the message event time is before year 2022, drop the message with event time unchanged.
    # If it's within year 2022, update the tag to "within_year_2022" and
    # update the message event time to Jan 1st 2022.
    # Otherwise, (exclusively after year 2022), update the tag to
    # "after_year_2022" and update the
    # message event time to Jan 1st 2023.

    january_first_2022 = datetime.datetime.fromtimestamp(1640995200)
    january_first_2023 = datetime.datetime.fromtimestamp(1672531200)


    async def my_handler(keys: list[str], datum: Datum) -> Messages:
        val = datum.value
        event_time = datum.event_time
        messages = Messages()

        if event_time < january_first_2022:
            logging.info("Got event time:%s, it is before 2022, so dropping", event_time)
            messages.append(Message.to_drop(event_time))
        elif event_time < january_first_2023:
            logging.info(
                "Got event time:%s, it is within year 2022, so forwarding to within_year_2022",
                event_time,
            )
            messages.append(
                Message(value=val, event_time=january_first_2022,
                        tags=["within_year_2022"])
            )
        else:
            logging.info(
                "Got event time:%s, it is after year 2022, so forwarding to
                after_year_2022", event_time
            )
            messages.append(Message(value=val, event_time=january_first_2023,
                            tags=["after_year_2022"]))

        return messages


    if __name__ == "__main__":
        grpc_server = SourceTransformAsyncServer(my_handler)
        grpc_server.start()
    """

    def __init__(
        self,
        source_transform_instance: SourceTransformAsyncCallable,
        sock_path=SOURCE_TRANSFORMER_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=SOURCE_TRANSFORMER_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Asynchronous Map Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
        mapper_instance: The mapper instance to be used for Map UDF
        sock_path: The UNIX socket path to be used for the server
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                     defaults to 4 and max capped at 16
        """
        self.sock_path = f"unix://{sock_path}"
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.max_message_size = max_message_size
        self.server_info_file = server_info_file

        self.source_transform_instance = source_transform_instance

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]
        self.servicer = SourceTransformAsyncServicer(handler=source_transform_instance)

    def start(self) -> None:
        """
        Starter function for the Async server class, need a separate caller
        so that all the async coroutines can be started from a single context
        """
        aiorun.run(self.aexec(), use_uvloop=True)

    async def aexec(self) -> None:
        """
        Starts the Async gRPC server on the given UNIX socket with
        given max threads.
        """

        # As the server is async, we need to create a new server instance in the
        # same thread as the event loop so that all the async calls are made in the
        # same context

        server_new = grpc.aio.server(options=self._server_options)
        server_new.add_insecure_port(self.sock_path)
        transform_pb2_grpc.add_SourceTransformServicer_to_server(self.servicer, server_new)

        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[
            ContainerType.Sourcetransformer
        ]

        # Start the async server
        await start_async_server(
            server_async=server_new,
            sock_path=self.sock_path,
            max_threads=self.max_threads,
            cleanup_coroutines=list(),
            server_info_file=self.server_info_file,
            server_info=serv_info,
        )



import asyncio
from collections.abc import AsyncIterable

from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2

from pynumaflow._constants import _LOGGER, STREAM_EOF, ERR_UDF_EXCEPTION_STRING
from pynumaflow.proto.sourcetransformer import transform_pb2, transform_pb2_grpc
from pynumaflow.shared.asynciter import NonBlockingIterator
from pynumaflow.shared.server import handle_async_error
from pynumaflow.sourcetransformer import Datum
from pynumaflow.sourcetransformer._dtypes import SourceTransformAsyncCallable
from pynumaflow.types import NumaflowServicerContext


class SourceTransformAsyncServicer(transform_pb2_grpc.SourceTransformServicer):
    """
    This class is used to create a new grpc SourceTransform Async Servicer instance.
    It implements the SourceTransformServicer interface from the proto
    transform_pb2_grpc.py file.
    Provides the functionality for the required rpc methods.
    """

    def __init__(
        self,
        handler: SourceTransformAsyncCallable,
    ):
        self.background_tasks = set()
        self.__transform_handler: SourceTransformAsyncCallable = handler

    async def SourceTransformFn(
        self,
        request_iterator: AsyncIterable[transform_pb2.SourceTransformRequest],
        context: NumaflowServicerContext,
    ) -> AsyncIterable[transform_pb2.SourceTransformResponse]:
        """
        Applies a transform function to a SourceTransformRequest stream
        The pascal case function name comes from the proto transform_pb2_grpc.py file.
        """
        try:
            # The first message to be received should be a valid handshake
            req = await request_iterator.__anext__()
            # check if it is a valid handshake req
            if not (req.handshake and req.handshake.sot):
                raise Exception("SourceTransformFn: expected handshake message")
            yield transform_pb2.SourceTransformResponse(
                handshake=transform_pb2.Handshake(sot=True),
            )

            # result queue to stream messages from the user code back to the client
            global_result_queue = NonBlockingIterator()

            # reader task to process the input task and invoke the required tasks
            producer = asyncio.create_task(
                self._process_inputs(request_iterator, global_result_queue)
            )

            # keep reading on result queue and send messages back
            consumer = global_result_queue.read_iterator()
            async for msg in consumer:
                # If the message is an exception, we raise the exception
                if isinstance(msg, BaseException):
                    await handle_async_error(context, msg, ERR_UDF_EXCEPTION_STRING)
                    return
                # Send window response back to the client
                else:
                    yield msg
            # wait for the producer task to complete
            await producer
        except BaseException as e:
            _LOGGER.critical("SourceTransformFnError, re-raising the error", exc_info=True)
            await handle_async_error(context, e, ERR_UDF_EXCEPTION_STRING)
            return

    async def _process_inputs(
        self,
        request_iterator: AsyncIterable[transform_pb2.SourceTransformRequest],
        result_queue: NonBlockingIterator,
    ):
        """
        Utility function for processing incoming SourceTransformRequest
        """
        try:
            # for each incoming request, create a background task to execute the
            # UDF code
            async for req in request_iterator:
                msg_task = asyncio.create_task(self._invoke_transform(req, result_queue))
                # save a reference to a set to store active tasks
                self.background_tasks.add(msg_task)
                msg_task.add_done_callback(self.background_tasks.discard)

            # Wait for all tasks to complete concurrently
            await asyncio.gather(*self.background_tasks)

            # send an EOF to result queue to indicate that all tasks have completed
            await result_queue.put(STREAM_EOF)

        except BaseException:
            _LOGGER.critical("SourceTransformFnError Error, re-raising the error", exc_info=True)

    async def _invoke_transform(
        self, request: transform_pb2.SourceTransformRequest, result_queue: NonBlockingIterator
    ):
        """
        Invokes the user defined function.
        """
        try:
            datum = Datum(
                keys=list(request.request.keys),
                value=request.request.value,
                event_time=request.request.event_time.ToDatetime(),
                watermark=request.request.watermark.ToDatetime(),
                headers=dict(request.request.headers),
            )
            msgs = await self.__transform_handler(list(request.request.keys), datum)
            results = []
            for msg in msgs:
                event_time_timestamp = _timestamp_pb2.Timestamp()
                event_time_timestamp.FromDatetime(dt=msg.event_time)
                results.append(
                    transform_pb2.SourceTransformResponse.Result(
                        keys=list(msg.keys),
                        value=msg.value,
                        tags=msg.tags,
                        event_time=event_time_timestamp,
                    )
                )
            await result_queue.put(
                transform_pb2.SourceTransformResponse(results=results, id=request.request.id)
            )
        except BaseException as err:
            _LOGGER.critical("SourceTransformFnError handler error", exc_info=True)
            await result_queue.put(err)

    async def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> transform_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto transform_pb2_grpc.py file.
        """
        return transform_pb2.ReadyResponse(ready=True)



import logging
import os
from dataclasses import dataclass, field
from enum import Enum
from importlib.metadata import version
from typing import TypeVar

from pynumaflow import setup_logging

_LOGGER = setup_logging(__name__)
if os.getenv("PYTHONDEBUG"):
    _LOGGER.setLevel(logging.DEBUG)

# Need to keep consistent with all SDKs and client
EOF = "U+005C__END__"

# Env variables to be passed in the info server metadata.
# These need to be accessed in the client using the same key.
# Format - (key, env_var)
METADATA_ENVS = [("CPU_LIMIT", "NUMAFLOW_CPU_LIMIT")]

# Metadata keys

# MAP_MODE_KEY is used in the server info object's metadata to indicate which map mode is enabled
MAP_MODE_KEY = "MAP_MODE"
# MULTIPROC_KEY is the field used to indicate that Multiproc map mode is enabled
# The value contains the number of servers spawned.
MULTIPROC_KEY = "MULTIPROC"

SI = TypeVar("SI", bound="ServerInfo")


class Protocol(str, Enum):
    """
    Enumerate grpc server connection protocol.
    """

    UDS = "uds"
    TCP = "tcp"


class Language(str, Enum):
    """
    Enumerate Numaflow SDK language.
    """

    PYTHON = "python"


class MapMode(str, Enum):
    """
    Enumerate Map Mode to be enabled
    """

    UnaryMap = "unary-map"
    StreamMap = "stream-map"
    BatchMap = "batch-map"


class ContainerType(str, Enum):
    """
    Enumerate Numaflow container type.
    """

    Sourcer = "sourcer"
    Sourcetransformer = "sourcetransformer"
    Sinker = "sinker"
    Mapper = "mapper"
    Reducer = "reducer"
    Reducestreamer = "reducestreamer"
    Sessionreducer = "sessionreducer"
    Sideinput = "sideinput"
    Fbsinker = "fb-sinker"
    Accumulator = "accumulator"


# Minimum version of Numaflow required by the current SDK version
# To update this value, please follow the instructions for MINIMUM_NUMAFLOW_VERSION in
# https://github.com/numaproj/numaflow-rs/blob/main/src/shared.rs
MINIMUM_NUMAFLOW_VERSION = {
    ContainerType.Sourcer: "1.4.0-z",
    ContainerType.Sourcetransformer: "1.4.0-z",
    ContainerType.Sinker: "1.4.0-z",
    ContainerType.Mapper: "1.4.0-z",
    ContainerType.Reducer: "1.4.0-z",
    ContainerType.Reducestreamer: "1.4.0-z",
    ContainerType.Sessionreducer: "1.4.0-z",
    ContainerType.Sideinput: "1.4.0-z",
    ContainerType.Fbsinker: "1.4.0-z",
    ContainerType.Accumulator: "1.5.0-z",
}


@dataclass
class ServerInfo:
    """
    ServerInfo is used for the gRPC server to provide the information such as protocol,
    sdk version, language, metadata to the client.
    Args:
        protocol: Protocol to use (UDS or TCP)
        language: Language used by the server, in our case, it's Python
        minimum_numaflow_version: lower bound for the supported Numaflow version
        version: Numaflow sdk version used by the server
        metadata: Any additional information to be provided (env vars)
    """

    protocol: Protocol
    language: Language
    minimum_numaflow_version: str
    version: str
    metadata: dict = field(default_factory=dict)

    @classmethod
    def get_default_server_info(cls) -> SI:
        serv_info = ServerInfo(
            protocol=Protocol.UDS,
            language=Language.PYTHON,
            # on the platform end, if minimum_numaflow_version is empty,
            # by default, the version compatibility check is skipped.
            minimum_numaflow_version="",
            version=get_sdk_version(),
            metadata=dict(),
        )
        return serv_info


def get_sdk_version() -> str:
    """
    Return the pynumaflow SDK version
    """
    try:
        return version("pynumaflow")
    except Exception as e:
        # Adding this to handle the case for local test/CI where pynumaflow
        # will not be installed as a package
        _LOGGER.error("Could not read SDK version %r", e, exc_info=True)
        return ""



from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class SideInputResponse(_message.Message):
    __slots__ = ("value", "no_broadcast")
    VALUE_FIELD_NUMBER: _ClassVar[int]
    NO_BROADCAST_FIELD_NUMBER: _ClassVar[int]
    value: bytes
    no_broadcast: bool
    def __init__(self, value: _Optional[bytes] = ..., no_broadcast: bool = ...) -> None: ...

class ReadyResponse(_message.Message):
    __slots__ = ("ready",)
    READY_FIELD_NUMBER: _ClassVar[int]
    ready: bool
    def __init__(self, ready: bool = ...) -> None: ...



from collections.abc import Iterable


from pynumaflow._constants import _LOGGER, STREAM_EOF
from pynumaflow.proto.sinker import sink_pb2_grpc, sink_pb2
from pynumaflow.shared.server import exit_on_error
from pynumaflow.shared.synciter import SyncIterator
from pynumaflow.shared.thread_with_return import ThreadWithReturnValue
from pynumaflow.sinker._dtypes import SinkSyncCallable
from pynumaflow.sinker.servicer.utils import (
    datum_from_sink_req,
    _create_read_handshake_response,
    build_sink_resp_results,
)
from pynumaflow.types import NumaflowServicerContext


class SyncSinkServicer(sink_pb2_grpc.SinkServicer):
    """
    This class is used to create a new grpc Sink servicer instance.
    It implements the SinkServicer interface from the proto sink.proto file.
    Provides the functionality for the required rpc methods.
    """

    def __init__(self, handler: SinkSyncCallable):
        self.handler: SinkSyncCallable = handler

    def SinkFn(
        self, request_iterator: Iterable[sink_pb2.SinkRequest], context: NumaflowServicerContext
    ) -> Iterable[sink_pb2.SinkResponse]:
        """
        Applies a sink function to datum elements.
        """

        try:
            # The first message to be received should be a valid handshake
            req = next(request_iterator)
            # check if it is a valid handshake req
            if not (req.handshake and req.handshake.sot):
                raise Exception("SinkFn: expected handshake message")
            yield _create_read_handshake_response()
            # cur_task is used to track the thread processing
            # the current batch of messages.
            cur_task = None
            # Use a queue backed to handle request batches
            req_queue = SyncIterator()

            # iterate of the incoming messages ot the sink
            for d in request_iterator:
                # if we do not have any active thread currently processing the batch
                # we need to create one and call the User function for processing the same.
                if cur_task is None:
                    # Use a queue to handle request batches
                    req_queue = SyncIterator()
                    cur_task = ThreadWithReturnValue(
                        target=self._invoke_sink, args=(req_queue, context)
                    )
                    cur_task.start()

                # when we have end of transmission message, we need to stop the processing the
                # current batch and wait for the next batch of messages.
                # We will also wait for the current task to finish processing the current batch.
                # We mark the current task as None to indicate that we are
                # ready to process the next batch.
                if d.status and d.status.eot:
                    req_queue.put(STREAM_EOF)
                    ret = cur_task.join()
                    yield sink_pb2.SinkResponse(results=ret)
                    # send EOT after each finishing sink responses
                    yield sink_pb2.SinkResponse(status=sink_pb2.TransmissionStatus(eot=True))
                    cur_task = None
                    continue

                # if we have a valid message, we will add it to the request queue for processing.
                datum = datum_from_sink_req(d)
                req_queue.put(datum)

            if cur_task:
                cur_task.join()

        except BaseException as err:
            # Handle exceptions
            err_msg = f"UDSinkError: {repr(err)}"
            _LOGGER.critical(err_msg, exc_info=True)
            exit_on_error(context, err_msg)
            return

    def _invoke_sink(self, request_queue: SyncIterator, context: NumaflowServicerContext):
        try:
            # Invoke the handler function with the request queue
            rspns = self.handler(request_queue.read_iterator())
            return build_sink_resp_results(rspns)
        except BaseException as err:
            err_msg = f"UDSinkError: {repr(err)}"
            _LOGGER.critical(err_msg, exc_info=True)
            exit_on_error(context, err_msg)
            raise err

    def IsReady(self, request, context: NumaflowServicerContext) -> sink_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        """
        return sink_pb2.ReadyResponse(ready=True)



import os
from abc import ABCMeta, abstractmethod
from collections.abc import Iterator, AsyncIterator
from dataclasses import dataclass
from datetime import datetime
from typing import Callable, Optional

from pynumaflow.shared.asynciter import NonBlockingIterator


@dataclass(init=False)
class Offset:
    """
    Args:
        offset: the offset of the datum.
        partition_id: partition_id indicates which partition of the source the datum belongs to.
    """

    __slots__ = ("_offset", "_partition_id")

    _offset: bytes
    _partition_id: int

    def __init__(self, offset: bytes, partition_id: int):
        self._offset = offset
        self._partition_id = partition_id

    @classmethod
    def offset_with_default_partition_id(cls, offset: bytes):
        """
        Returns an Offset object with the given offset and default partition id.
        """
        return Offset(offset=offset, partition_id=get_default_partitions()[0])

    @property
    def as_dict(self):
        return {"offset": self._offset, "partition_id": self._partition_id}

    @property
    def offset(self) -> bytes:
        return self._offset

    @property
    def partition_id(self) -> int:
        return self._partition_id


@dataclass(init=False)
class Message:
    """
    Basic datatype for data passing to the next vertex/vertices.

    Args:
        payload: data in bytes
        offset: the offset of the datum.
        event_time: event time of the message, usually extracted from the payload.
        keys: []string keys for vertex (optional)
        headers: dict of headers for the message (optional)
    """

    __slots__ = ("_payload", "_offset", "_event_time", "_keys", "_headers")

    _payload: bytes
    _offset: Offset
    _event_time: datetime
    _keys: list[str]
    _headers: dict[str, str]

    def __init__(
        self,
        payload: bytes,
        offset: Offset,
        event_time: datetime,
        keys: list[str] = None,
        headers: Optional[dict[str, str]] = None,
    ):
        """
        Creates a Message object to send value to a vertex.
        """
        self._payload = payload
        self._offset = offset
        self._event_time = event_time
        self._keys = keys or []
        self._headers = headers or {}

    @property
    def payload(self) -> bytes:
        return self._payload

    @property
    def keys(self) -> list[str]:
        return self._keys

    @property
    def offset(self) -> Offset:
        return self._offset

    @property
    def event_time(self) -> datetime:
        return self._event_time

    @property
    def headers(self) -> dict[str, str]:
        return self._headers


@dataclass(init=False)
class ReadRequest:
    """
    Class to define the request for reading datum stream from user defined source.
    Args:
        num_records: the number of records to read.
        timeout_in_ms: the request timeout in milliseconds.
    >>> # Example usage
    >>> from pynumaflow.sourcer import ReadRequest
    >>> read_request = ReadRequest(num_records=10, timeout_in_ms=1000)
    """

    __slots__ = ("_num_records", "_timeout_in_ms")

    _num_records: int
    _timeout_in_ms: int

    def __init__(
        self,
        num_records: int,
        timeout_in_ms: int,
    ):
        if not isinstance(num_records, int):
            raise TypeError(f"Wrong data type: {type(num_records)} for ReadRequest.num_records")
        self._num_records = num_records
        if not isinstance(timeout_in_ms, int):
            raise TypeError(f"Wrong data type: {type(timeout_in_ms)} for ReadRequest.timeout_in_ms")
        self._timeout_in_ms = timeout_in_ms

    @property
    def num_records(self) -> int:
        """Returns the num_records of the request"""
        return self._num_records

    @property
    def timeout_in_ms(self) -> int:
        """Returns the timeout_in_ms of the request."""
        return self._timeout_in_ms


@dataclass(init=False)
class AckRequest:
    """
    Class for defining the request for acknowledging datum.
    It takes a list of offsets that need to be acknowledged.
    Args:
        offsets: the offsets to be acknowledged.
    >>> # Example usage
    >>> from pynumaflow.sourcer import AckRequest, Offset
    >>> offset_val = Offset(offset=b"123", partition_id=0)
    >>> ack_request = AckRequest(offsets=[offset_val, offset_val])
    """

    __slots__ = ("_offsets",)
    _offsets: list[Offset]

    def __init__(self, offsets: list[Offset]):
        self._offsets = offsets

    @property
    def offsets(self) -> list[Offset]:
        """Returns the offsets to be acknowledged."""
        return self._offsets


@dataclass
class NackRequest:
    """
    Class for defining the request for negatively acknowledging an offset.
    It takes a list of offsets that need to be negatively acknowledged on the source.
    Args:
        offsets: the offsets to be negatively acknowledged.
    >>> # Example usage
    >>> from pynumaflow.sourcer import NackRequest, Offset
    >>> offset_val = Offset(offset=b"123", partition_id=0)
    >>> nack_request = NackRequest(offsets=[offset_val, offset_val])
    """

    __slots__ = ("_offsets",)
    _offsets: list[Offset]

    def __init__(self, offsets: list[Offset]):
        self._offsets = offsets

    @property
    def offsets(self) -> list[Offset]:
        """Returns the offsets to be negatively acknowledged."""
        return self._offsets


@dataclass(init=False)
class PendingResponse:
    """
    PendingResponse is the response for the pending request.
    It indicates the number of pending records at the user defined source.
    A negative count indicates that the pending information is not available.
    Args:
        count: the number of pending records.
    """

    __slots__ = ("_count",)
    _count: int

    def __init__(self, count: int):
        if not isinstance(count, int):
            raise TypeError(f"Wrong data type: {type(count)} for Pending.count")
        self._count = count

    @property
    def count(self) -> int:
        """Returns the count of pending records"""
        return self._count


@dataclass(init=False)
class PartitionsResponse:
    """
    PartitionsResponse is the response for the partition request.
    It indicates the number of partitions at the user defined source.
    A negative count indicates that the partition information is not available.
    Args:
        count: the number of partitions.
    """

    __slots__ = ("_partitions",)
    _partitions: list[int]

    def __init__(self, partitions: list[int]):
        if not isinstance(partitions, list):
            raise TypeError(f"Wrong data type: {type(partitions)} for Partition.partitions")
        self._partitions = partitions

    @property
    def partitions(self) -> list[int]:
        """Returns the list of partitions"""
        return self._partitions


class Sourcer(metaclass=ABCMeta):
    """
    Provides an interface to write a Sourcer
    which will be exposed over an gRPC server.

    Args:

    """

    def __call__(self, *args, **kwargs):
        """
        Allow to call handler function directly if class instance is sent
        """
        return self.handler(*args, **kwargs)

    @abstractmethod
    async def read_handler(self, datum: ReadRequest, output: NonBlockingIterator):
        """
        Implement this handler function which implements the SourceReadCallable interface.
        read_handler is used to read the data from the source and send the data forward
        for each read request we process num_records and increment the read_idx to indicate that
        the message has been read and the same is added to the ack set
        """
        pass

    @abstractmethod
    def ack_handler(self, ack_request: AckRequest):
        """
        The ack handler is used to acknowledge the offsets that have been read
        """
        pass

    @abstractmethod
    def nack_handler(self, nack_request: NackRequest):
        """
        The nack handler is used to negatively acknowledge the offsets on the source
        """
        pass

    @abstractmethod
    def pending_handler(self) -> PendingResponse:
        """
        The simple source always returns zero to indicate there is no pending record.
        """
        pass

    @abstractmethod
    def partitions_handler(self) -> PartitionsResponse:
        """
        The simple source always returns zero to indicate there is no pending record.
        """
        pass


# Create default partition id from the environment variable "NUMAFLOW_REPLICA"
DefaultPartitionId = int(os.getenv("NUMAFLOW_REPLICA", "0"))
SourceReadCallable = Callable[[ReadRequest], Iterator[Message]]
AsyncSourceReadCallable = Callable[[ReadRequest], AsyncIterator[Message]]
SourceAckCallable = Callable[[AckRequest], None]
SourceCallable = Sourcer


def get_default_partitions() -> list[int]:
    """
    Returns the default partition ids.
    """
    return [DefaultPartitionId]



from abc import ABCMeta, abstractmethod
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from datetime import datetime
from typing import TypeVar, Callable, Union, Optional
from collections.abc import AsyncIterable
from warnings import warn

from pynumaflow._constants import DROP

M = TypeVar("M", bound="Message")
Ms = TypeVar("Ms", bound="Messages")


@dataclass(init=False)
class Message:
    """
    Basic datatype for data passing to the next vertex/vertices.

    Args:
        value: data in bytes
        keys: []string keys for vertex (optional)
        tags: []string tags for conditional forwarding (optional)
    """

    __slots__ = ("_value", "_keys", "_tags")

    _value: bytes
    _keys: list[str]
    _tags: list[str]

    def __init__(self, value: bytes, keys: list[str] = None, tags: list[str] = None):
        """
        Creates a Message object to send value to a vertex.
        """
        self._keys = keys or []
        self._tags = tags or []
        self._value = value or b""

    # returns the Message Object which will be dropped
    @classmethod
    def to_drop(cls: type[M]) -> M:
        return cls(b"", None, [DROP])

    @property
    def value(self) -> bytes:
        return self._value

    @property
    def keys(self) -> list[str]:
        return self._keys

    @property
    def tags(self) -> list[str]:
        return self._tags


class Messages(Sequence[M]):
    """
    Class to define a list of Message objects.

    Args:
        messages: list of Message objects.
    """

    __slots__ = ("_messages",)

    def __init__(self, *messages: M):
        self._messages = list(messages) or []

    def __str__(self) -> str:
        return str(self._messages)

    def __repr__(self) -> str:
        return str(self)

    def __len__(self) -> int:
        return len(self._messages)

    def __iter__(self) -> Iterator[M]:
        return iter(self._messages)

    def __getitem__(self, index: int) -> M:
        if isinstance(index, slice):
            raise TypeError("Slicing is not supported for Messages")
        return self._messages[index]

    def append(self, message: Message) -> None:
        self._messages.append(message)

    def items(self) -> list[Message]:
        warn(
            "Using items is deprecated and will be removed in v0.5. "
            "Iterate or index the Messages object instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._messages


@dataclass(init=False)
class Datum:
    """
    Class to define the important information for the event.
    Args:
        keys: the keys of the event.
        value: the payload of the event.
        event_time: the event time of the event.
        watermark: the watermark of the event.
        headers: the headers of the event.

    >>> # Example usage
    >>> from pynumaflow.mapstreamer import Datum
    >>> from datetime import datetime, timezone
    >>> payload = bytes("test_mock_message", encoding="utf-8")
    >>> t1 = datetime.fromtimestamp(1662998400, timezone.utc)
    >>> t2 = datetime.fromtimestamp(1662998460, timezone.utc)
    >>> msg_headers = {"key1": "value1", "key2": "value2"}
    >>> d = Datum(
    ...       keys=["test_key"],
    ...       value=payload,
    ...       event_time=t1,
    ...       watermark=t2,
    ...       headers=msg_headers,
    ...    )
    """

    __slots__ = ("_keys", "_value", "_event_time", "_watermark", "_headers")

    _keys: list[str]
    _value: bytes
    _event_time: datetime
    _watermark: datetime
    _headers: dict[str, str]

    def __init__(
        self,
        keys: list[str],
        value: bytes,
        event_time: datetime,
        watermark: datetime,
        headers: Optional[dict[str, str]] = None,
    ):
        self._keys = keys or list()
        self._value = value or b""
        if not isinstance(event_time, datetime):
            raise TypeError(f"Wrong data type: {type(event_time)} for Datum.event_time")
        self._event_time = event_time
        if not isinstance(watermark, datetime):
            raise TypeError(f"Wrong data type: {type(watermark)} for Datum.watermark")
        self._watermark = watermark
        self._headers = headers or {}

    @property
    def keys(self) -> list[str]:
        """Returns the keys of the event"""
        return self._keys

    @property
    def value(self) -> bytes:
        """Returns the value of the event."""
        return self._value

    @property
    def event_time(self) -> datetime:
        """Returns the event time of the event."""
        return self._event_time

    @property
    def watermark(self) -> datetime:
        """Returns the watermark of the event."""
        return self._watermark

    @property
    def headers(self) -> dict[str, str]:
        """Returns the headers of the event."""
        return self._headers.copy()


class MapStreamer(metaclass=ABCMeta):
    """
    Provides an interface to write a Map Streamer
    which will be exposed over a gRPC server.

    Args:

    """

    def __call__(self, *args, **kwargs):
        """
        Allow to call handler function directly if class instance is sent
        """
        return self.handler(*args, **kwargs)

    @abstractmethod
    async def handler(self, keys: list[str], datum: Datum) -> AsyncIterable[Message]:
        """
        Implement this handler function which implements the MapSyncCallable interface.
        """
        pass


MapStreamAsyncCallable = Callable[[list[str], Datum], AsyncIterable[Message]]
MapStreamCallable = Union[MapStreamer, MapStreamAsyncCallable]


class MapStreamError(Exception):
    """To Raise an error while executing a MapStream call"""



from collections.abc import AsyncIterable
from typing import Union

from google.protobuf import empty_pb2 as _empty_pb2

from pynumaflow._constants import _LOGGER, ERR_UDF_EXCEPTION_STRING
from pynumaflow.proto.reducer import reduce_pb2, reduce_pb2_grpc
from pynumaflow.reducer._dtypes import (
    Datum,
    ReduceAsyncCallable,
    _ReduceBuilderClass,
    ReduceRequest,
    WindowOperation,
)
from pynumaflow.reducer.servicer.task_manager import TaskManager
from pynumaflow.shared.server import handle_async_error
from pynumaflow.types import NumaflowServicerContext


async def datum_generator(
    request_iterator: AsyncIterable[reduce_pb2.ReduceRequest],
) -> AsyncIterable[ReduceRequest]:
    """
    This function is used to create an async generator
    from the gRPC request iterator.
    It yields a ReduceRequest instance for each request received which is then
    forwarded to the task manager.
    """
    async for d in request_iterator:
        reduce_request = ReduceRequest(
            operation=d.operation.event,
            windows=d.operation.windows,
            payload=Datum(
                keys=list(d.payload.keys),
                value=d.payload.value,
                event_time=d.payload.event_time.ToDatetime(),
                watermark=d.payload.watermark.ToDatetime(),
                headers=dict(d.payload.headers),
            ),
        )
        yield reduce_request


class AsyncReduceServicer(reduce_pb2_grpc.ReduceServicer):
    """
    This class is used to create a new grpc Reduce servicer instance.
    It implements the SyncMapServicer interface from the proto reduce.proto file.
    Provides the functionality for the required rpc methods.
    """

    def __init__(
        self,
        handler: Union[ReduceAsyncCallable, _ReduceBuilderClass],
    ):
        # The Reduce handler can be a function or a builder class instance.
        self.__reduce_handler: Union[ReduceAsyncCallable, _ReduceBuilderClass] = handler

    async def ReduceFn(
        self,
        request_iterator: AsyncIterable[reduce_pb2.ReduceRequest],
        context: NumaflowServicerContext,
    ) -> reduce_pb2.ReduceResponse:
        """
        Applies a reduce function to a datum stream.
        The pascal case function name comes from the proto reduce_pb2_grpc.py file.

        The lifecycle of the reduce operation is managed by the task manager. Which is created
        for each new grpc call.

        Whenever a new request is received, the task manager invokes the required
        operation based on the request type.
        1) If the request is for a new keyed window, a new task is created.
           If the request is for an existing keyed window, the data is
           appended to the task.

        2) Once the request iterator is exhausted, the task manager sends EOF to all
        the tasks, to signal them to stop reading the data from their respective iterators.

        3) The task manager then waits for the tasks to complete and
        yields the response from each task

        """
        # Create an async iterator from the request iterator
        datum_iterator = datum_generator(request_iterator=request_iterator)

        # Create a task manager instance
        # The task manager is used to manage lifecycle of the tasks
        # required for the reduce operation.
        task_manager = TaskManager(handler=self.__reduce_handler, context=context)

        # Start iterating through the request iterator and create tasks
        # based on the operation type received.
        try:
            async for request in datum_iterator:
                # check whether the request is an open or append operation
                if request.operation is int(WindowOperation.OPEN):
                    # create a new task for the open operation
                    # and inserts the request data into the task
                    await task_manager.create_task(request)
                elif request.operation is int(WindowOperation.APPEND):
                    # append the task data to the existing task
                    # if the task does not exist, it will create a new task
                    await task_manager.append_task(request)
        except BaseException as e:
            _LOGGER.critical("Reduce Error", exc_info=True)
            # Send a context abort signal for the rpc, this is required for numa container to get
            # the correct grpc error
            await handle_async_error(context, e, ERR_UDF_EXCEPTION_STRING)

        # send EOF to all the tasks once the request iterator is exhausted
        # This will signal the tasks to stop reading the data on their
        # respective iterators.
        await task_manager.stream_send_eof()

        # Get the list of tasks from the task manager
        res = task_manager.get_tasks()
        try:
            # iterate through the tasks and yield the response
            # from each of them once the task is completed.
            for task in res:
                fut = task.future
                await fut

                # For each message in the task result, yield the response
                if fut.result():
                    for msg in fut.result():
                        yield reduce_pb2.ReduceResponse(result=msg, window=task.window)

            # For each window processed by the ReduceFn send an EOF response
            # We send one EOF per window
            current_window = task_manager.get_unique_windows()
            for window in current_window.values():
                # yield the EOF response once the task is completed for a keyed window
                yield reduce_pb2.ReduceResponse(window=window, EOF=True)
        except BaseException as e:
            _LOGGER.critical("Reduce Error", exc_info=True)
            # Send a context abort signal for the rpc, this is required for numa container to get
            # the correct grpc error
            await handle_async_error(context, e, ERR_UDF_EXCEPTION_STRING)

    async def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> reduce_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto reduce_pb2_grpc.py file.
        """
        return reduce_pb2.ReadyResponse(ready=True)



import datetime

from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ReduceRequest(_message.Message):
    __slots__ = ("payload", "operation")
    class WindowOperation(_message.Message):
        __slots__ = ("event", "windows")
        class Event(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            OPEN: _ClassVar[ReduceRequest.WindowOperation.Event]
            CLOSE: _ClassVar[ReduceRequest.WindowOperation.Event]
            APPEND: _ClassVar[ReduceRequest.WindowOperation.Event]
        OPEN: ReduceRequest.WindowOperation.Event
        CLOSE: ReduceRequest.WindowOperation.Event
        APPEND: ReduceRequest.WindowOperation.Event
        EVENT_FIELD_NUMBER: _ClassVar[int]
        WINDOWS_FIELD_NUMBER: _ClassVar[int]
        event: ReduceRequest.WindowOperation.Event
        windows: _containers.RepeatedCompositeFieldContainer[Window]
        def __init__(self, event: _Optional[_Union[ReduceRequest.WindowOperation.Event, str]] = ..., windows: _Optional[_Iterable[_Union[Window, _Mapping]]] = ...) -> None: ...
    class Payload(_message.Message):
        __slots__ = ("keys", "value", "event_time", "watermark", "headers")
        class HeadersEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        KEYS_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        EVENT_TIME_FIELD_NUMBER: _ClassVar[int]
        WATERMARK_FIELD_NUMBER: _ClassVar[int]
        HEADERS_FIELD_NUMBER: _ClassVar[int]
        keys: _containers.RepeatedScalarFieldContainer[str]
        value: bytes
        event_time: _timestamp_pb2.Timestamp
        watermark: _timestamp_pb2.Timestamp
        headers: _containers.ScalarMap[str, str]
        def __init__(self, keys: _Optional[_Iterable[str]] = ..., value: _Optional[bytes] = ..., event_time: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., watermark: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., headers: _Optional[_Mapping[str, str]] = ...) -> None: ...
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    payload: ReduceRequest.Payload
    operation: ReduceRequest.WindowOperation
    def __init__(self, payload: _Optional[_Union[ReduceRequest.Payload, _Mapping]] = ..., operation: _Optional[_Union[ReduceRequest.WindowOperation, _Mapping]] = ...) -> None: ...

class Window(_message.Message):
    __slots__ = ("start", "end", "slot")
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    SLOT_FIELD_NUMBER: _ClassVar[int]
    start: _timestamp_pb2.Timestamp
    end: _timestamp_pb2.Timestamp
    slot: str
    def __init__(self, start: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., end: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., slot: _Optional[str] = ...) -> None: ...

class ReduceResponse(_message.Message):
    __slots__ = ("result", "window", "EOF")
    class Result(_message.Message):
        __slots__ = ("keys", "value", "tags")
        KEYS_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        TAGS_FIELD_NUMBER: _ClassVar[int]
        keys: _containers.RepeatedScalarFieldContainer[str]
        value: bytes
        tags: _containers.RepeatedScalarFieldContainer[str]
        def __init__(self, keys: _Optional[_Iterable[str]] = ..., value: _Optional[bytes] = ..., tags: _Optional[_Iterable[str]] = ...) -> None: ...
    RESULT_FIELD_NUMBER: _ClassVar[int]
    WINDOW_FIELD_NUMBER: _ClassVar[int]
    EOF_FIELD_NUMBER: _ClassVar[int]
    result: ReduceResponse.Result
    window: Window
    EOF: bool
    def __init__(self, result: _Optional[_Union[ReduceResponse.Result, _Mapping]] = ..., window: _Optional[_Union[Window, _Mapping]] = ..., EOF: bool = ...) -> None: ...

class ReadyResponse(_message.Message):
    __slots__ = ("ready",)
    READY_FIELD_NUMBER: _ClassVar[int]
    ready: bool
    def __init__(self, ready: bool = ...) -> None: ...



import asyncio
from datetime import datetime, timezone
from typing import Union
from collections.abc import AsyncIterable

import grpc

from pynumaflow.exceptions import UDFError
from pynumaflow.proto.reducer import reduce_pb2
from pynumaflow.shared.asynciter import NonBlockingIterator
from pynumaflow._constants import (
    STREAM_EOF,
    DELIMITER,
    _LOGGER,
)
from pynumaflow.reducer._dtypes import (
    IntervalWindow,
    Metadata,
    ReduceResult,
    Datum,
    _ReduceBuilderClass,
    ReduceAsyncCallable,
    ReduceWindow,
)
from pynumaflow.shared.server import exit_on_error
from pynumaflow.types import NumaflowServicerContext


def build_unique_key_name(keys, window):
    """
    Builds a unique key name for the given keys and window.
    The key name is used to identify the Reduce task.
    The format is: start_time:end_time:key1:key2:...
    """
    return f"{window.start.ToMilliseconds()}:{window.end.ToMilliseconds()}:{DELIMITER.join(keys)}"


def build_window_hash(window):
    """
    Builds a hash for the given window.
    The hash is used to identify the Reduce Window
    The format is: start_time:end_time
    """
    return f"{window.start.ToMilliseconds()}:{window.end.ToMilliseconds()}"


class TaskManager:
    """
    TaskManager is responsible for managing the Reduce tasks.
    It is created whenever a new reduce operation is requested.
    """

    def __init__(
        self,
        handler: Union[ReduceAsyncCallable, _ReduceBuilderClass],
        context: NumaflowServicerContext,
    ):
        # A dictionary to store the task information
        self.tasks = {}
        # Collection for storing strong references to all running tasks.
        # Event loop only keeps a weak reference, which can cause it to
        # get lost during execution.
        self.background_tasks = set()
        # Handler for the reduce operation
        self.__reduce_handler = handler
        # Servicer context from grpc, required to abort if error occurs
        self.context = context

    def get_tasks(self):
        """
        Returns the list of reduce tasks that are
        currently being processed
        """
        return self.tasks.values()

    def get_unique_windows(self):
        """
        Returns the unique windows that are currently being processed
        """
        # Dict to store unique windows
        windows = dict()
        # Iterate over all the tasks and add the windows
        for task in self.tasks.values():
            window_hash = build_window_hash(task.window)
            window_found = windows.get(window_hash, None)
            # if window not seen yet, add to the dict
            if not window_found:
                windows[window_hash] = task.window
        return windows

    async def stream_send_eof(self):
        """
        Sends EOF to input streams of all the Reduce
        tasks that are currently being processed.
        This is called when the input grpc stream is closed.
        """
        for unified_key in self.tasks:
            await self.tasks[unified_key].iterator.put(STREAM_EOF)

    async def create_task(self, req):
        """
        Creates a new reduce task for the given request.
        Based on the request we compute a unique key, and then
        it creates a new task or appends the request to the existing task.
        """
        # if len of windows in request != 1, raise error
        if len(req.windows) != 1:
            raise UDFError("reduce create operation error: invalid number of windows")

        d = req.payload
        keys = d.keys
        unified_key = build_unique_key_name(keys, req.windows[0])
        result = self.tasks.get(unified_key, None)

        # If the task does not exist, create a new task
        if not result:
            niter = NonBlockingIterator()
            riter = niter.read_iterator()
            task = asyncio.create_task(self.__invoke_reduce(keys, riter, req.windows[0]))
            # Save a reference to the result of this function, to avoid a
            # task disappearing mid-execution.
            self.background_tasks.add(task)
            task.add_done_callback(lambda t: self.background_tasks.remove(t))
            result = ReduceResult(task, niter, keys, req.windows[0])

            # Save the result of the reduce operation to the task list
            self.tasks[unified_key] = result

        # Put the request in the iterator
        await result.iterator.put(d)

    async def append_task(self, req):
        """
        Appends the request to the existing window reduce task.
        If the task does not exist, create it.
        """
        if len(req.windows) != 1:
            raise UDFError("reduce create operation error: invalid number of windows")
        d = req.payload
        keys = d.keys
        unified_key = build_unique_key_name(keys, req.windows[0])
        result = self.tasks.get(unified_key, None)
        if not result:
            await self.create_task(req)
        else:
            await result.iterator.put(d)

    async def __invoke_reduce(
        self, keys: list[str], request_iterator: AsyncIterable[Datum], window: ReduceWindow
    ):
        """
        Invokes the UDF reduce handler with the given keys,
        request iterator, and window. Returns the result of the
        reduce operation.
        """
        new_instance = self.__reduce_handler

        # Convert the window to a datetime object
        start_dt = datetime.fromtimestamp(int(window.start.ToMilliseconds()) / 1e3, timezone.utc)
        end_dt = datetime.fromtimestamp(int(window.end.ToMilliseconds()) / 1e3, timezone.utc)
        interval_window = IntervalWindow(start_dt, end_dt)
        md = Metadata(interval_window=interval_window)
        # If the reduce handler is a class instance, create a new instance of it.
        # It is required for a new key to be processed by a
        # new instance of the reducer for a given window
        # Otherwise the function handler can be called directly
        if isinstance(self.__reduce_handler, _ReduceBuilderClass):
            new_instance = self.__reduce_handler.create()
        try:
            msgs = await new_instance(keys, request_iterator, md)
        except BaseException as err:
            _LOGGER.critical("UDFError, re-raising the error", exc_info=True)
            # Send a context abort signal for the rpc, this is required for numa container to get
            # the correct grpc error
            await asyncio.gather(
                self.context.abort(grpc.StatusCode.UNKNOWN, details=repr(err)),
                return_exceptions=True,
            )
            exit_on_error(err=repr(err), parent=False, context=self.context, update_context=False)
            return

        datum_responses = []
        for msg in msgs:
            datum_responses.append(
                reduce_pb2.ReduceResponse.Result(keys=msg.keys, value=msg.value, tags=msg.tags)
            )

        return datum_responses



import asyncio
import contextlib
import io
import multiprocessing
import os
import socket
import traceback

from google.protobuf import any_pb2
from google.rpc import code_pb2, status_pb2, error_details_pb2
from grpc_status import rpc_status
from abc import ABCMeta, abstractmethod
from collections.abc import Iterator
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

import grpc
import psutil

from pynumaflow._constants import (
    _LOGGER,
    MULTIPROC_MAP_SOCK_ADDR,
    UDFType,
)
from pynumaflow.exceptions import SocketError
from pynumaflow.info.server import write as info_server_write, get_metadata_env
from pynumaflow.info.types import (
    ServerInfo,
    METADATA_ENVS,
    MULTIPROC_KEY,
)
from pynumaflow.proto.mapper import map_pb2_grpc
from pynumaflow.proto.sideinput import sideinput_pb2_grpc
from pynumaflow.proto.sinker import sink_pb2_grpc
from pynumaflow.proto.sourcer import source_pb2_grpc
from pynumaflow.proto.sourcetransformer import transform_pb2_grpc
from pynumaflow.types import NumaflowServicerContext


class NumaflowServer(metaclass=ABCMeta):
    """
    Provides an interface to write a Numaflow Server
    which will be exposed over gRPC.
    """

    @abstractmethod
    def start(self):
        """
        Start the gRPC server
        """
        pass


def sync_server_start(
    servicer,
    bind_address: str,
    max_threads: int,
    server_info_file: str,
    server_options=None,
    server_info: Optional[ServerInfo] = None,
    udf_type: str = UDFType.Map,
):
    """
    Utility function to start a sync grpc server instance.
    """
    if server_info is None:
        # Add the server information to the server info file
        # if not provided
        server_info = ServerInfo.get_default_server_info()

    # Run a sync server instance
    _run_server(
        servicer=servicer,
        bind_address=bind_address,
        threads_per_proc=max_threads,
        server_options=server_options,
        udf_type=udf_type,
        server_info_file=server_info_file,
        server_info=server_info,
    )


def _run_server(
    servicer,
    bind_address: str,
    threads_per_proc,
    server_options,
    udf_type: str,
    server_info_file: Optional[str] = None,
    server_info: Optional[ServerInfo] = None,
) -> None:
    """
    Starts the Synchronous server instance on the given UNIX socket
    with given max threads. Wait for the server to terminate.
    """
    server = grpc.server(
        ThreadPoolExecutor(
            max_workers=threads_per_proc,
        ),
        options=server_options,
    )

    # add the correct servicer to the server based on the UDF type
    if udf_type == UDFType.Map:
        map_pb2_grpc.add_MapServicer_to_server(servicer, server)
    elif udf_type == UDFType.Sink:
        sink_pb2_grpc.add_SinkServicer_to_server(servicer, server)
    elif udf_type == UDFType.SourceTransformer:
        transform_pb2_grpc.add_SourceTransformServicer_to_server(servicer, server)
    elif udf_type == UDFType.Source:
        source_pb2_grpc.add_SourceServicer_to_server(servicer, server)
    elif udf_type == UDFType.SideInput:
        sideinput_pb2_grpc.add_SideInputServicer_to_server(servicer, server)

    # bind the server to the UDS/TCP socket
    server.add_insecure_port(bind_address)
    # start the gRPC server
    server.start()
    # Add the server information to the server info file if provided
    if server_info and server_info_file:
        info_server_write(server_info=server_info, info_file=server_info_file)

    _LOGGER.info("GRPC Server listening on: %s %d", bind_address, os.getpid())
    server.wait_for_termination()


def start_multiproc_server(
    max_threads: int,
    servicer,
    process_count: int,
    server_info_file: str,
    server_info: Optional[ServerInfo] = None,
    server_options=None,
    udf_type: str = UDFType.Map,
):
    """
    Start N grpc servers in different processes where N = The number of CPUs or the
    value of the env var NUM_CPU_MULTIPROC defined by the user. The max value
    is set to 2 * CPU count.
    Each server will be bound to a different UDS socket, and we will create equal number of
    workers to handle each server.
    On the client side there will be same number of connections as the number of servers.
    """

    _LOGGER.info(
        "Starting new Multiproc server with num_procs: %s, num_threads per proc: %s",
        process_count,
        max_threads,
    )
    workers = []
    for idx in range(process_count):
        # bind address is the UDS sock for each server to  bind to, it is in the format
        # unix:///var/run/numaflow/multiproc#serv_num.sock
        # -> unix:///var/run/numaflow/multiproc0.sock
        bind_address = f"unix://{MULTIPROC_MAP_SOCK_ADDR}{idx}.sock"
        _LOGGER.info("Starting server on: %s", bind_address)
        # NOTE: It is imperative that the worker subprocesses be forked before
        # any gRPC servers start up. See
        # https://github.com/grpc/grpc/issues/16001 for more details.
        worker = multiprocessing.Process(
            target=_run_server,
            args=(servicer, bind_address, max_threads, server_options, udf_type),
        )
        worker.start()
        workers.append(worker)

    if server_info is None:
        server_info = ServerInfo.get_default_server_info()
    server_info.metadata = get_metadata_env(envs=METADATA_ENVS)
    # Add the MULTIPROC metadata using the number of servers to use
    server_info.metadata[MULTIPROC_KEY] = str(process_count)
    info_server_write(server_info=server_info, info_file=server_info_file)

    for worker in workers:
        worker.join()


async def start_async_server(
    server_async: grpc.aio.Server,
    sock_path: str,
    max_threads: int,
    cleanup_coroutines: list,
    server_info_file: str,
    server_info: Optional[ServerInfo] = None,
):
    """
    Starts the Async server instance on the given UNIX socket with given max threads.
    Add the server graceful shutdown coroutine to the cleanup_coroutines list.
    Wait for the server to terminate.
    """
    await server_async.start()

    if server_info is None:
        # Create the server info file if not provided
        server_info = ServerInfo.get_default_server_info()
    # Add the server information to the server info file
    info_server_write(server_info=server_info, info_file=server_info_file)

    # Log the server start
    _LOGGER.info(
        "Async GRPC Server listening on: %s with max threads: %s",
        sock_path,
        max_threads,
    )

    async def server_graceful_shutdown():
        """
        Shuts down the server with 5 seconds of grace period. During the
        grace period, the server won't accept new connections and allow
        existing RPCs to continue within the grace period.
        """
        _LOGGER.info("Starting graceful shutdown...")
        await server_async.stop(5)

    cleanup_coroutines.append(server_graceful_shutdown())
    await server_async.wait_for_termination()


@contextlib.contextmanager
def _reserve_port(port_num: int) -> Iterator[int]:
    """Find and reserve a port for all subprocesses to use."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    if sock.getsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR) == 0:
        raise SocketError("Failed to set SO_REUSEADDR.")
    try:
        sock.bind(("", port_num))
        yield sock.getsockname()[1]
    finally:
        sock.close()


def check_instance(instance, callable_type) -> bool:
    """
    Check if the given instance is of the given callable_type.
    """
    try:
        if not isinstance(instance, callable_type):
            return False
        else:
            return True
    except Exception as e:
        _LOGGER.error(e)
        return False


def get_grpc_status(err: str):
    """
    Create a grpc status object with the error details.
    """
    details = any_pb2.Any()
    details.Pack(
        error_details_pb2.DebugInfo(
            detail="\n".join(traceback.format_stack()),
        )
    )

    status = status_pb2.Status(code=code_pb2.INTERNAL, message=err, details=[details])
    return rpc_status.to_status(status)


def exit_on_error(
    context: NumaflowServicerContext, err: str, parent: bool = False, update_context=True
):
    """
    Exit the current/parent process on an error.

    Args:
        context (NumaflowServicerContext): The gRPC context.
        err (str): The error message.
        parent (bool, optional): Whether this is the parent process.
            Defaults to False.
        update_context(bool, optional) : Is there a need to update
            the context with the error codes
    """
    if update_context:
        # Create a status object with the error details
        grpc_status = get_grpc_status(err)

        context.set_code(grpc.StatusCode.INTERNAL)
        context.set_details(err)
        context.set_trailing_metadata(grpc_status.trailing_metadata)

    p = psutil.Process(os.getpid())
    # If the parent flag is true, we exit from the parent process
    # Use this for Multiproc right now to exit from the parent fork
    if parent:
        p = psutil.Process(os.getppid())
    _LOGGER.info("Killing process: Got exception %s", err)
    p.kill()


def update_context_err(context: NumaflowServicerContext, e: BaseException, err_msg: str):
    """
    Update the context with the error and log the exception.
    """
    trace = get_exception_traceback_str(e)
    _LOGGER.critical(trace)
    _LOGGER.critical(e.__str__())

    grpc_status = get_grpc_status(err_msg)

    context.set_code(grpc.StatusCode.INTERNAL)
    context.set_details(err_msg)
    context.set_trailing_metadata(grpc_status.trailing_metadata)


def get_exception_traceback_str(exc) -> str:
    file = io.StringIO()
    traceback.print_exception(exc, value=exc, tb=exc.__traceback__, file=file)
    return file.getvalue().rstrip()


async def handle_async_error(
    context: NumaflowServicerContext, exception: BaseException, exception_type: str
):
    """
    Handle exceptions for async servers by updating the context and exiting.
    """
    err_msg = f"{exception_type}: {repr(exception)}"
    update_context_err(context, exception, err_msg)
    await asyncio.gather(
        context.abort(grpc.StatusCode.INTERNAL, details=err_msg), return_exceptions=True
    )
    exit_on_error(err=err_msg, parent=False, context=context, update_context=False)
